#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from logging import Handler, LogRecord

from qtpy.QtCore import qInstallMessageHandler, Slot
from qtpy.QtWidgets import QWidget, QComboBox

from MDANSE.MLogging import LOG, FMT

from MDANSE_GUI.Tabs.GeneralTab import GeneralTab
from MDANSE_GUI.Tabs.Layouts.SinglePanel import SinglePanel
from MDANSE_GUI.Session.LocalSession import LocalSession
from MDANSE_GUI.Tabs.Visualisers.TextInfo import TextInfo
from MDANSE_GUI.Widgets.DefaultCombobox import highlight_default_value


log_tab_label = """MDANSE_GUI message log.

This tab will display the general logging messages of the graphical interface.
You can adjust the logging level using the combo box below.
"""


class GuiLogHandler(Handler):

    def __init__(self, *args, **kwargs):
        self._visualiser = None
        super().__init__(*args, **kwargs)
        self.setFormatter(FMT)

    def add_visualiser(self, new_visualiser):
        self._visualiser = new_visualiser

    def emit(self, record: LogRecord):
        if self._visualiser is not None:
            message = self.formatter.format(record)
            if "WARNING" in message:
                message = f'<span style="color:orange;">{message}</span>'
            elif "ERROR" in message or "CRITICAL" in message:
                message = f'<span style="color:red;">{message}</span>'
            else:
                message = f"<span>{message}</span>"
            self._visualiser.append_text(message)


class LoggingTab(GeneralTab):
    """The tab for tracking the progress of running jobs."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._extra_handler = None
        self._visualiser.toHtml()
        qInstallMessageHandler(self.log_qt_handler)
        self._loglevel_combo = QComboBox(self._core)
        self._loglevel_combo.addItems(["INFO", "DEBUG", "WARNING", "ERROR", "CRITICAL"])
        self._loglevel_combo.setCurrentText("INFO")
        self._loglevel_combo.currentTextChanged.connect(self.change_log_level)
        highlight_default_value(self._loglevel_combo)
        self._core._ub_layout.addWidget(self._loglevel_combo)

    @Slot(str)
    def change_log_level(self, new_level: str):
        if self._extra_handler is None:
            return
        try:
            self._extra_handler.setLevel(new_level)
        except:
            LOG.error(f"Could not set GuiLogHandler to log level {new_level}")
        else:
            self._visualiser.append_text(
                f"<b>=== Log level changed to {new_level} ===</b>"
            )

    def add_handler(self, new_handler):
        try:
            current_level = self._loglevel_combo.currentText()
        except:
            current_level = "INFO"
        self._extra_handler = new_handler
        self._extra_handler.add_visualiser(self._visualiser)
        self.change_log_level(current_level)

    def log_qt_handler(self, m_type, m_context, m_text):
        self._visualiser.append_text(f"Qt log message (type={m_type})=" + m_text)

    @classmethod
    def standard_instance(cls):
        the_tab = cls(
            window,
            name="Logger",
            visualiser=TextInfo(),
            layout=SinglePanel,
            label_text=log_tab_label,
        )
        return the_tab

    @classmethod
    def gui_instance(
        cls,
        parent: QWidget,
        name: str,
        session: LocalSession,
        settings,
        logger,
        **kwargs,
    ):
        the_tab = cls(
            parent,
            name=name,
            session=session,
            settings=settings,
            logger=logger,
            visualiser=TextInfo(),
            layout=SinglePanel,
            label_text=log_tab_label,
        )
        return the_tab


if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication, QMainWindow, QVBoxLayout

    app = QApplication(sys.argv)
    window = QMainWindow()
    the_tab = LoggingTab(
        window,
        name="RunningJobs",
        visualiser=TextInfo(),
        layout=SinglePanel,
        label_text=log_tab_label,
    )
    window.setCentralWidget(the_tab._core)
    window.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from qtpy.QtWidgets import (
    QDialog,
    QPushButton,
    QVBoxLayout,
    QComboBox,
    QTextEdit,
)
from qtpy.QtCore import Signal, Slot, QObject

from MDANSE.Framework.Jobs.IJob import IJob


sample_styles = {
    "blank_stylesheet": "",
    "initial_stylesheet": """QWidget
{   border-radius: 1px;
    border-style: 1nset;
    border-width: 0px;
    border-color: rgb(30,250,250);
    font: bold 14px;
    padding: 2px;
}""",
    "dark_stylesheet": """QWidget
{   background-color:rgb(40,10,80);
    border-radius: 6px;
    border-style: outset;
    border-width: 2px;
    border-color: rgb(10,20,200);
    font: bold 14px;
    color: rgb(10,200,10);
    padding: 3px;
}""",
    "hippie_stylesheet": """QWidget
{   background-color: qlineargradient(x1:0,y1:0,x2:1,y2:1,stop: 0 #aa1163, stop: 1 #23bb54);
    border-radius: 10px;
    border-style: outset;
    border-width: 3px;
    border-color: rgb(10,200,50);
    font: bold 14px;
    color: rgb(10,200,230);
    padding: 3px;
}""",
}


class StyleDatabase(QObject):
    stored_style = Signal(str)
    stored_labels = Signal(list)

    def __init__(self, parent: Optional["QObject"] = ...) -> None:
        super().__init__(parent)
        self._styles = sample_styles

    @Slot(list)
    def addStyleString(self, style: list[str]):
        label = style[0]
        text = style[1]
        self._styles[label] = text

    @Slot(str)
    def returnStyleString(self, label: str) -> str:
        try:
            temp = self._styles[label]
        except KeyError:
            return None
        else:
            self.stored_style.emit(temp)
            return temp

    @Slot()
    def showStoredLabels(self) -> list[str]:
        temp = [str(x) for x in self._styles.keys()]
        self.stored_labels.emit(temp)
        return temp


class StyleDialog(QDialog):
    new_style = Signal(str)
    icon_swap = Signal(bool)

    def __init__(self, *args, converter: IJob = "Dummy", **kwargs):
        super().__init__(*args, **kwargs)

        self._database = None

        layout = QVBoxLayout(self)
        self.setLayout(layout)
        self._selector = QComboBox(self)
        self._display = QTextEdit(self)
        self._confirm = QPushButton("Apply", self)
        self._confirm.clicked.connect(self.commitChanges)

        for x in [self._selector, self._display, self._confirm]:
            layout.addWidget(x)

    def connectStyleDatabase(self, dbase: StyleDatabase):
        self._database = dbase
        labels = dbase.showStoredLabels()
        self._selector.clear()
        [self._selector.addItem(lab) for lab in labels]
        dbase.stored_style.connect(self._display.setText)
        self._selector.currentTextChanged.connect(dbase.returnStyleString)
        self._selector.setCurrentIndex(0)

    def commitChanges(self):
        text = self._display.document().toPlainText()
        self.new_style.emit(text)
        label = self._selector.currentText()
        if "dark" in label:
            self.icon_swap.emit(True)
        else:
            self.icon_swap.emit(False)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, List

from matplotlib.pyplot import style as mpl_style
from matplotlib import rcParams

from MDANSE.MLogging import LOG
from MDANSE.Core.SubclassFactory import SubclassFactory

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext


class Plotter(metaclass=SubclassFactory):

    def __init__(self) -> None:
        self._figure = None
        self._axes = []
        self._initial_values = [0.0, 0.0]
        self._slider_values = [0.0, 0.0]
        self._number_of_sliders = 2
        self._value_reset_needed = True
        self._toolbar = None
        self._slider_reference = None

    def request_slider_values(self):
        if self._slider_reference is None:
            return
        self._slider_reference.collect_values()

    def clear(self, figure: "Figure" = None):
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            return
        target.clear()

    def slider_labels(self) -> List[str]:
        return ["Slider 1", "Slider 2"]

    def slider_limits(self) -> List[str]:
        return self._number_of_sliders * [[-1.0, 1.0, 0.01]]

    def sliders_coupled(self) -> bool:
        return False

    def get_figure(self, figure: "Figure" = None):
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            LOG.error(f"PlottingContext can't plot to {target}")
            return
        target.clear()
        return target

    def apply_settings(self, plotting_context: "PlottingContext"):
        if plotting_context.set_axes() is None:
            LOG.debug("Axis check failed.")
            return

    def enable_slider(self, allow_slider: bool = True):
        if allow_slider:
            self._slider_reference.setEnabled(True)
            self._slider_reference.blockSignals(False)
        else:
            self._slider_reference.setEnabled(False)
            self._slider_reference.blockSignals(True)

    def handle_slider(self, new_value: List[float]):
        self._slider_values = new_value

    def plot(
        self,
        plotting_context: "PlottingContext",
        figure: "Figure" = None,
        colours=None,
        update_only=False,
        toolbar=None,
    ):
        target = self.get_figure(figure)
        if target is None:
            return
        if toolbar is not None:
            self._toolbar = toolbar
        axes = target.add_subplot(111)
        self._axes = [axes]
        self.apply_settings(plotting_context, colours)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from qtpy.QtWidgets import QCheckBox
from qtpy.QtCore import Qt

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class BooleanWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        default_option = self._configurator.default
        field = QCheckBox(self._base)
        field.setTristate(False)
        field.setChecked(default_option)
        field.stateChanged.connect(self.updateValue)
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "A single logical value that can be True of False"
        field.setToolTip(tooltip_text)
        self._field = field
        self._layout.addWidget(field)
        self.default_labels()
        self.update_labels()
        self.updateValue()

    def default_labels(self):
        """Each Widget should have a default tooltip and label,
        which will be set in this method, unless specific
        values are provided in the settings of the job that
        is being configured."""
        if self._label_text == "":
            self._label_text = "BooleanWidget"
        if self._tooltip == "":
            self._tooltip = "A single logical value that can be True of False"
        self._field.setToolTip(self._tooltip)

    def configure_using_default(self):
        """No need to anything for Boolean"""

    def get_widget_value(self):
        return self._field.checkState() == Qt.CheckState.Checked



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import os
from functools import partial

from qtpy.QtCore import Slot
from qtpy.QtWidgets import QWidget, QFileDialog

from MDANSE.Framework.InputData.HDFTrajectoryInputData import HDFTrajectoryInputData

from MDANSE_GUI.Tabs.GeneralTab import GeneralTab
from MDANSE_GUI.Tabs.Layouts.MultiPanel import MultiPanel
from MDANSE_GUI.Session.LocalSession import LocalSession
from MDANSE_GUI.Tabs.Models.GeneralModel import GeneralModel
from MDANSE_GUI.Tabs.Views.TrajectoryView import TrajectoryView
from MDANSE_GUI.Tabs.Visualisers.TrajectoryInfo import TrajectoryInfo
from MDANSE_GUI.Tabs.Visualisers.View3D import View3D
from MDANSE_GUI.MolecularViewer.MolecularViewer import MolecularViewer


label_text = """Here you can load the .mdt files.
They are MD trajectories in HDF5 format
created by one of the MDANSE converters.
Any trajectory you select will be visualised
in the 3D view window on the right side.
The animation of the MD trajectory will
allow you to verify if the contents 
of the trajectory are what you expected.
"""


class TrajectoryTab(GeneralTab):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._core.add_button("Load .MDT Trajectories", self.load_trajectories)

    @Slot()
    def load_trajectories(self):
        fnames = QFileDialog.getOpenFileNames(
            self._core,
            "Load an MD trajectory",
            self.get_path("trajectory"),
            "HDF5 files, MDANSE or H5MD format (*.mdt *.h5);;H5MD files (*.h5);;All files (*)",
        )
        for fname in fnames[0]:
            self.load_trajectory(fname)
            last_path, _ = os.path.split(fname)
        if fnames[0]:
            self.set_path("trajectory", last_path)
            self._session.save()

    @Slot(str)
    def load_trajectory(self, fname: str):
        if len(fname) > 0:
            _, short_name = os.path.split(fname)
            try:
                data = HDFTrajectoryInputData(fname)
            except Exception as e:
                self._core.error.emit(repr(e))
            else:
                self._core._model.append_object(((fname, data), short_name))

    @classmethod
    def standard_instance(cls):
        the_tab = cls(
            window,
            name="Trajectories",
            session=LocalSession(),
            model=GeneralModel(),
            view=TrajectoryView(),
            visualiser=View3D(MolecularViewer()),
            layout=partial(MultiPanel, left_panels=[TrajectoryInfo()]),
            label_text=label_text,
        )
        return the_tab

    @classmethod
    def gui_instance(
        cls,
        parent: QWidget,
        name: str,
        session: LocalSession,
        settings,
        logger,
        **kwargs,
    ):
        the_tab = cls(
            parent,
            name=name,
            session=session,
            settings=settings,
            logger=logger,
            model=kwargs.get("model", GeneralModel()),
            view=TrajectoryView(),
            visualiser=View3D(MolecularViewer()),
            layout=partial(MultiPanel, left_panels=[TrajectoryInfo()]),
            label_text=label_text,
        )
        return the_tab


if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication, QMainWindow

    app = QApplication(sys.argv)
    window = QMainWindow()
    the_tab = TrajectoryTab.standard_instance()
    window.setCentralWidget(the_tab._core)
    window.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import copy

from qtpy.QtWidgets import (
    QDialog,
    QPushButton,
    QFrame,
    QGridLayout,
    QLabel,
    QApplication,
    QLineEdit,
    QListView,
)
from qtpy.QtCore import (
    Signal,
    Slot,
    QModelIndex,
)
from qtpy.QtGui import (
    QStandardItem,
    QStandardItemModel,
    QDoubleValidator,
)

from MDANSE.Framework.Units import UNITS_MANAGER

from MDANSE_GUI.Widgets.GeneralWidgets import InputVariable, InputDialog


class UnitVisualiser(QFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        layout = QGridLayout(self)
        self.setLayout(layout)

        components = ["factor", "kg", "m", "s", "K", "mol", "A", "cd", "rad", "sr"]
        self.lines = {}

        validator = QDoubleValidator()

        for n, item in enumerate(components):
            label = QLabel(item, self)
            entry = QLineEdit("0", self)
            entry.setValidator(validator)
            layout.addWidget(label, n, 0)
            layout.addWidget(entry, n, 1)
            self.lines[item] = entry

    @Slot(dict)
    def newValues(self, val_dict: dict):
        for key, value in val_dict.items():
            entry = self.lines[key]
            entry.setReadOnly(False)
            entry.setText(str(value))
            entry.setReadOnly(True)


class UnitModel(QStandardItemModel):
    unit_info = Signal(dict)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._defaultUnits = copy.deepcopy(UNITS_MANAGER.units)

        self.populateUnitList()

    def populateUnitList(self):
        for unit in self._defaultUnits:
            item = QStandardItem(unit)
            item.setData(unit)
            self.appendRow(item)

    @Slot(QModelIndex)
    def passValue(self, index: QModelIndex):
        label = self.itemFromIndex(index)
        unit = UNITS_MANAGER.units[label.text()]
        info = {}
        info["factor"] = unit._factor
        info["kg"] = unit._dimension[0]
        info["m"] = unit._dimension[1]
        info["s"] = unit._dimension[2]
        info["K"] = unit._dimension[3]
        info["mol"] = unit._dimension[4]
        info["A"] = unit._dimension[5]
        info["cd"] = unit._dimension[6]
        info["rad"] = unit._dimension[7]
        info["sr"] = unit._dimension[8]
        self.unit_info.emit(info)

    @Slot(dict)
    def addNewUnit(self, info: dict):
        unit_name = info["name"]
        if UNITS_MANAGER.has_unit(unit_name):
            return None
        else:
            UNITS_MANAGER.add_unit(
                unit_name,
                *[
                    info[key]
                    for key in [
                        "factor",
                        "kg",
                        "m",
                        "s",
                        "K",
                        "mol",
                        "A",
                        "cd",
                        "rad",
                        "sr",
                    ]
                ],
            )
            UNITS_MANAGER.save()
            item = QStandardItem(unit_name)
            item.setData(unit_name)
            self.appendRow(item)


class UnitsEditor(QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle("MDANSE Physical Units Editor")

        self.model = UnitModel()

        layout = QGridLayout(self)
        self.setLayout(layout)

        self.view = QListView(self)
        self.view.setModel(self.model)

        self.visualiser = UnitVisualiser(self)

        self.view.clicked.connect(self.model.passValue)
        self.model.unit_info.connect(self.visualiser.newValues)

        new_unit_button = QPushButton("New unit", self)
        new_unit_button.clicked.connect(self.newUnitDialog)

        layout.addWidget(self.view, 0, 0)
        layout.addWidget(self.visualiser, 0, 1)
        layout.addWidget(new_unit_button, 1, 0)

    @Slot()
    def newUnitDialog(self):
        dialog_variables = [
            InputVariable(
                input_dict={
                    "keyval": "name",
                    "format": str,
                    "label": "Unit name",
                    "tooltip": "Type the name of the new physical unit here.",
                    "values": [""],
                }
            ),
            InputVariable(
                input_dict={
                    "keyval": "factor",
                    "format": float,
                    "label": "Scaling factor",
                    "tooltip": "Type the scaling factor of the new unit here.",
                    "values": [1.0],
                }
            ),
        ] + [
            InputVariable(
                input_dict={
                    "keyval": tag,
                    "format": int,
                    "label": tag,
                    "tooltip": "Type the power to which the component should be raised.",
                    "values": [0],
                }
            )
            for tag in ["kg", "m", "s", "K", "mol", "A", "cd", "rad", "sr"]
        ]
        nu_dialog = InputDialog(fields=dialog_variables)
        nu_dialog.got_values.connect(self.model.addNewUnit)
        nu_dialog.show()
        result = nu_dialog.exec()


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    root = UnitsEditor()
    root.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from functools import partial

from qtpy.QtWidgets import QWidget

from MDANSE_GUI.Tabs.GeneralTab import GeneralTab
from MDANSE_GUI.Tabs.Layouts.MultiPanel import MultiPanel
from MDANSE_GUI.Session.LocalSession import LocalSession
from MDANSE_GUI.Tabs.Visualisers.TextInfo import TextInfo
from MDANSE_GUI.Tabs.Models.JobHolder import JobHolder
from MDANSE_GUI.Tabs.Views.RunTable import RunTable
from MDANSE_GUI.Tabs.Visualisers.JobLogInfo import JobLogInfo


run_tab_label = """This table shows the status of jobs
which have been started in your session.
You can check which jobs were successful,
and if they failed, you can see the details
of the error message.
"""


class RunTab(GeneralTab):
    """The tab for tracking the progress of running jobs."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def standard_instance(cls):
        the_tab = cls(
            window,
            name="RunMonitor",
            model=JobHolder(),
            view=RunTable(),
            visualiser=TextInfo(
                header="MDANSE Jobs",
                footer="Look up our Read The Docs page:"
                + "https://mdanse.readthedocs.io/en/protos/",
            ),
            layout=partial(
                MultiPanel,
                right_panels=[JobLogInfo(header="MDANSE Logs")],
            ),
            label_text=run_tab_label,
        )
        return the_tab

    @classmethod
    def gui_instance(
        cls,
        parent: QWidget,
        name: str,
        session: LocalSession,
        settings,
        logger,
        **kwargs,
    ):
        the_tab = cls(
            parent,
            name=name,
            session=session,
            settings=settings,
            logger=logger,
            model=kwargs.get("model", JobHolder()),
            view=RunTable(),
            visualiser=TextInfo(
                header="MDANSE Jobs",
                footer="Look up our "
                + '<a href="https://mdanse.readthedocs.io/en/protos/">Read The Docs</a>'
                + " page.",
            ),
            layout=partial(
                MultiPanel,
                right_panels=[JobLogInfo(header="MDANSE Logs")],
            ),
            label_text=run_tab_label,
        )
        return the_tab


if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication, QMainWindow, QVBoxLayout

    app = QApplication(sys.argv)
    window = QMainWindow()
    the_tab = RunTab(
        window,
        name="RunningJobs",
        model=JobHolder(),
        view=RunTable(),
        visualiser=TextInfo(),
        layout=partial(MultiPanel, right_panels=[JobLogInfo()]),
        label_text=run_tab_label,
    )
    window.setCentralWidget(the_tab._core)
    window.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import hashlib
import os

from qtpy.QtGui import QStandardItemModel, QStandardItem
from qtpy.QtCore import QObject, Slot

from MDANSE.MLogging import LOG

from MDANSE.Framework.InputData import InputFileData
from MDANSE.Framework.InputData.HDFTrajectoryInputData import HDFTrajectoryInputData

# class TrajectoryItem(QStandardItem):

#     def __init__(self, *args, trajectory = None):
#         super().__init__(*args)
#         self._trajectory = trajectory


# class TrajectoryHolder(QStandardItemModel):

#     def __init__(self, parent: QObject = None):
#         super().__init__(parent=parent)

#     @Slot(object)
#     def addItem(self, new_entry: InputFileData):
#         traj = TrajectoryItem(new_entry.basename, trajectory = new_entry)
#         self.appendRow([traj])

#     @Slot(object)
#     def acceptNewTrajectory(self, new_trajectory):


class FileObject:
    def __init__(self, hash_function=hashlib.sha256):
        self.filename = ""
        self.absolute_path = ""
        self.extension = ".*"
        self.relative_path = "."
        self.hash_function = hash_function
        self.hash = -1

    def setFilename(self, fname: str):
        abspath, filename = os.path.split(fname)
        self.extension = str(filename).split(".")[-1]
        self.filename = filename
        self.absolute_path = abspath
        self.hash = self.calculateHash(fname)

    def calculateHash(self, fname, chunk_size=1024 * 512):
        hash_object = self.hash_function()
        with open(fname, "br") as source:
            while chunk := source.read(chunk_size):
                hash_object.update(chunk)
        return hash_object.hexdigest()


class DataTreeItem(QStandardItem):
    def __init__(self, *args, **kwargs):
        new_kwargs = {}
        for key, val in kwargs.items():
            if key not in ["mdanse_tag"]:
                new_kwargs[key] = val
        super().__init__(*args, **new_kwargs)

        self.mdanse_tag = kwargs.get("mdanse_tag", "unknown")

        self.original_input = []
        self.processing_used = ""
        self.processing_parameters = {}

    def showHistory(self):
        history = ";".join(self.ancestors())
        return history

    def ancestors(self):
        parent = self.parent()
        if parent is None:
            return [self.mdanse_tag]
        else:
            return parent.ancestors() + [self.mdanse_tag]


class TrajectoryItem(DataTreeItem):
    def __init__(self, *args, **kwargs):
        new_kwargs = {}
        for key, val in kwargs.items():
            if key not in ["fname", "trajectory"]:
                new_kwargs[key] = val
        super().__init__(*args, **new_kwargs)

        self.filename = kwargs.get("fname", "NULL")
        self.trajectory = kwargs.get("trajectory", None)
        self.original_input = []
        self.processing_used = ""
        self.processing_parameters = {}
        self.file_info = FileObject()

        if self.filename != "NULL":
            self.file_info.setFilename(self.filename)
        self.setText(os.path.split(self.filename)[1])


class AnalysisItem(DataTreeItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class DataTreeModel(QStandardItemModel):
    def __init__(self, parent: QObject = None):
        super().__init__(parent=parent)
        self._trajectory_objects = {}

    @Slot(object)
    def addItem(self, new_entry: InputFileData):
        traj = TrajectoryItem(new_entry.basename, trajectory=new_entry)
        self.appendRow([traj])

    @Slot(str)
    def acceptNewTrajectory(self, fname: str):
        LOG.info(f"Received {fname}")
        data = HDFTrajectoryInputData(fname)
        item = TrajectoryItem(fname=fname, trajectory=data, mdanse_tag="hdf_trajectory")
        LOG.info("Created TrajectoryItem")
        self.appendRow(item)
        tkey = item.file_info.hash
        self._trajectory_objects[tkey] = data
        LOG.info("assigned new trajectory to the dictionary")



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from qtpy.QtCore import Slot, Qt, QTimer, QMutex
from qtpy.QtWidgets import (
    QWidget,
    QGridLayout,
    QSlider,
    QHBoxLayout,
    QSpinBox,
    QVBoxLayout,
    QPushButton,
    QStyle,
    QSizePolicy,
    QTreeView,
    QDoubleSpinBox,
    QColorDialog,
    QGroupBox,
    QCheckBox,
)
from qtpy.QtGui import (
    QPaintEvent,
    QPainter,
    QColor,
)

from MDANSE_GUI.Tabs.Views.Delegates import ColourPicker, RadiusSpinBox
from MDANSE_GUI.MolecularViewer.MolecularViewer import MolecularViewer

button_lookup = {
    "start": QStyle.StandardPixmap.SP_MediaSkipBackward,
    "back": QStyle.StandardPixmap.SP_MediaSeekBackward,
    "stop": QStyle.StandardPixmap.SP_MediaStop,
    "play": QStyle.StandardPixmap.SP_MediaPlay,
    "fwd": QStyle.StandardPixmap.SP_MediaSeekForward,
    "end": QStyle.StandardPixmap.SP_MediaSkipForward,
}

jogger_stylesheet = """QSpinBox {
  border: 1px solid #ABABAB;
  border-radius: 3px;
}

QSpinBox::down-button  {
  subcontrol-origin: margin;
  subcontrol-position: center left;
  image: url(:/icons/leftArrow.png);
}

QSpinBox::up-button  {
  subcontrol-origin: margin;
  subcontrol-position: center right;
  image: url(:/icons/rightArrow.png);
}"""

another_stylesheet = """QSpinBox {
  border: 1px solid #ABABAB;
  border-radius: 3px;
}

QSpinBox::up-button  {
  subcontrol-origin: margin;
  subcontrol-position: center left;
  image: url(:/icons/leftArrow.png);
  background-color: #ABABAB;
  left: 1px;
  height: 24px;
  width: 24px;
}

QSpinBox::down-button  {
  subcontrol-origin: margin;
  subcontrol-position: center right;
  image: url(:/icons/rightArrow.png);
  background-color: #ABABAB;
  right: 1px;
  height: 24px;
  width: 24px;
}
"""


class ViewerControls(QWidget):
    def __init__(self, *args, **kwargs):
        super(QWidget, self).__init__(*args, **kwargs)
        layout = QGridLayout(self)
        self._viewer = None
        self._buttons = {}
        self._delegates = {}
        self._animation_timer = QTimer()
        self._animation_timer.timeout.connect(self.advance_frame)
        self._mutex = QMutex()
        self._frame_step = 1
        self._time_per_frame = 80  # in ms
        self._frame_factor = 1  # just a scalar multiplication factor
        self._visibility = [True, True, True, True]
        self.createSlider()
        self.createButtons(Qt.Orientation.Horizontal)
        self.createSidePanel()
        self._bkg_dialog = QColorDialog()
        self._projection = True

    def createSlider(self):
        """This method creates a slider which illustrates the progress of the
        trajectory animation.
        """
        base = QWidget(self)
        base.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)
        hlay = QHBoxLayout(base)
        frame_slider = QSlider(base)
        frame_slider.setToolTip("Select a frame")
        frame_slider.setOrientation(Qt.Orientation.Horizontal)
        self._frame_slider = frame_slider
        hlay.addWidget(frame_slider)
        frame_selector = QSpinBox(base)
        # frame_selector.setStyleSheet(jogger_stylesheet)
        frame_selector.setValue(0)
        frame_selector.setMinimum(0)
        hlay.addWidget(frame_selector)
        frame_selector.valueChanged.connect(frame_slider.setValue)
        frame_slider.valueChanged.connect(frame_selector.setValue)
        self._frame_selector = frame_selector
        self.layout().addWidget(base, 2, 0, 1, 3)  # row, column, rowSpan, columnSpan

    def setViewer(self, viewer: MolecularViewer):
        self._viewer = viewer
        self.layout().addWidget(viewer, 0, 0, 2, 2)  # row, column, rowSpan, columnSpan
        self._frame_slider.valueChanged.connect(viewer.set_coordinates)
        viewer.new_max_frames.connect(self._frame_slider.setMaximum)
        viewer.new_max_frames.connect(self._frame_selector.setMaximum)
        viewer.new_max_frames.connect(self.stop_animation)
        # self._database.setViewer(viewer)
        # viewer.setDataModel(viewer._colour_manager)
        self._atom_details.setModel(viewer._colour_manager)
        for column_number in range(3):
            self._atom_details.resizeColumnToContents(column_number)
        viewer._colour_manager.new_atom_properties.connect(viewer.take_atom_properties)

    def createButtons(self, orientation: Qt.Orientation):
        """Create a bar with video player buttons for controlling the
        animation.
        """
        base = QWidget(self)
        base.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)
        if orientation == Qt.Orientation.Horizontal:
            layout = QHBoxLayout(base)
            self.layout().addWidget(
                base, 3, 0, 1, 3
            )  # row, column, rowSpan, columnSpan
        elif orientation == Qt.Orientation.Vertical:
            layout = QVBoxLayout(base)
            self.layout().addWidget(
                base, 0, 2, 2, 1
            )  # row, column, rowSpan, columnSpan
        for button_name, button_role in button_lookup.items():
            temp = QPushButton(base)
            icon = self.style().standardIcon(button_role)
            temp.setIcon(icon)
            layout.addWidget(temp)
            self._buttons[button_name] = temp
        self._buttons["stop"].clicked.connect(self.stop_animation)
        self._buttons["play"].clicked.connect(lambda: self.animate())
        self._buttons["fwd"].clicked.connect(lambda: self.animate(step_size=10))
        self._buttons["back"].clicked.connect(lambda: self.animate(step_size=-10))
        self._buttons["start"].clicked.connect(lambda: self.go_to_end(last_frame=False))
        self._buttons["end"].clicked.connect(lambda: self.go_to_end(last_frame=True))

    def createSidePanel(self):
        """Adds widgets for finer control of the playback"""
        base = QWidget(self)
        layout = QVBoxLayout(base)
        base.setLayout(layout)
        # colour changes
        wrapper0 = QGroupBox("Colour settings", base)
        layout0 = QHBoxLayout(wrapper0)
        bkg_button = QPushButton("Background", wrapper0)
        proj_button = QPushButton("Toggle projection", wrapper0)
        bkg_button.clicked.connect(self.set_background_colour)
        proj_button.clicked.connect(self.toggle_projection)
        layout0.addWidget(bkg_button)
        layout0.addWidget(proj_button)
        layout.addWidget(wrapper0)
        # the table of chemical elements
        wrapper1 = QGroupBox("Atom properties", base)
        layout1 = QHBoxLayout(wrapper1)
        self._atom_details = QTreeView(base)
        self._atom_details.setSizePolicy(
            QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum
        )
        self._delegates["colour"] = ColourPicker()
        self._delegates["radius"] = RadiusSpinBox()
        self._atom_details.setItemDelegateForColumn(1, self._delegates["colour"])
        self._atom_details.setItemDelegateForColumn(2, self._delegates["radius"])
        layout1.addWidget(self._atom_details)
        wrapper1.setLayout(layout1)
        layout.addWidget(wrapper1)
        # the widget selecting time per animation frame
        wrapper2 = QGroupBox("Time per frame (ms)", base)
        layout2 = QHBoxLayout(wrapper2)
        frame_time_selector = QSpinBox(base)
        frame_time_selector.setToolTip("Larger number means slower animation")
        frame_time_selector.setValue(5)
        frame_time_selector.setMinimum(1)
        frame_time_selector.setMaximum(5000)
        frame_time_selector.valueChanged.connect(self.setTimeStep)
        layout2.addWidget(frame_time_selector)
        wrapper2.setLayout(layout2)
        layout.addWidget(wrapper2)
        # the widget for frame skipping
        wrapper3 = QGroupBox("Show every N frames", base)
        layout3 = QHBoxLayout(wrapper3)
        frame_skip = QSpinBox(base)
        frame_skip.setToolTip("Animate every N frames in fast forward")
        frame_skip.setValue(1)
        frame_skip.setMinimum(1)
        frame_skip.setMaximum(5000)
        frame_skip.valueChanged.connect(self.setFrameSkip)
        layout3.addWidget(frame_skip)
        wrapper3.setLayout(layout3)
        layout.addWidget(wrapper3)
        # the widget for atom size scaling
        wrapper4 = QGroupBox("Atom size scaling", base)
        layout4 = QHBoxLayout(wrapper4)
        size_factor = QDoubleSpinBox(base)
        size_factor.setToolTip("Scaling factor for atom size")
        size_factor.setValue(0.8)
        size_factor.setMinimum(0.0)
        size_factor.setMaximum(50.0)
        size_factor.setSingleStep(0.05)
        size_factor.valueChanged.connect(self.setAtomSize)
        layout4.addWidget(size_factor)
        wrapper4.setLayout(layout4)
        layout.addWidget(wrapper4)
        wrapper5 = QGroupBox("Visible Objects", base)
        layout5 = QHBoxLayout(wrapper5)
        atoms_visible = QCheckBox("atoms", wrapper5)
        bonds_visible = QCheckBox("bonds", wrapper5)
        axes_visible = QCheckBox("axes", wrapper5)
        cell_visible = QCheckBox("cell", wrapper5)
        self._visibility_checkboxes = [
            atoms_visible,
            bonds_visible,
            axes_visible,
            cell_visible,
        ]
        for nw, box in enumerate(self._visibility_checkboxes):
            box.setTristate(False)
            box.setChecked(self._visibility[nw])
            box.stateChanged.connect(self.setVisibility)
            layout5.addWidget(box)
        layout.addWidget(wrapper5)
        # the database of atom types
        # self._database = TrajectoryAtomData()
        self.layout().addWidget(base, 0, 2, 2, 1)  # row, column, rowSpan, columnSpan

    @Slot()
    def set_background_colour(self):
        dialog = self._bkg_dialog
        if dialog.isVisible():
            if dialog.isMaximized():
                dialog.showMaximized()
            else:
                dialog.showNormal()
            dialog.activateWindow()
        else:
            dialog.show()
        dialog.exec()
        if dialog.result() == QColorDialog.DialogCode.Accepted:
            colour = dialog.currentColor()
            rgb = colour.red() / 255, colour.green() / 255, colour.blue() / 255
            self._viewer._renderer.SetBackground(rgb)

    @Slot()
    def toggle_projection(self):
        if self._projection:
            self._viewer._camera.SetParallelProjection(255)
        else:
            self._viewer._camera.SetParallelProjection(0)
        self._viewer.update_renderer()
        self._projection = not self._projection

    @Slot()
    def setVisibility(self):
        for nw, box in enumerate(self._visibility_checkboxes):
            self._visibility[nw] = box.isChecked()
        self._viewer._new_visibility(self._visibility)

    @Slot(int)
    def setTimeStep(self, new_value: int):
        self._time_per_frame = new_value
        self.animate()

    @Slot(int)
    def setFrameSkip(self, new_value: int):
        self._frame_factor = new_value
        self.animate()

    @Slot(float)
    def setAtomSize(self, new_value: float):
        if self._viewer is None:
            return
        self._viewer._new_scaling(new_value)

    @Slot(bool)
    def go_to_end(self, last_frame: bool = True):
        self.stop_animation()
        if last_frame:
            target = self._frame_selector.maximum()
        else:
            target = self._frame_selector.minimum()
        self._frame_selector.setValue(target)

    @Slot()
    def stop_animation(self):
        """Stop changing trajectory frames."""
        self._mutex.lock()
        if self._animation_timer.isActive():
            self._animation_timer.stop()
        self._mutex.unlock()

    def animate(self, step_size: int = 1):
        """Plays the animation as a movie.

        Keyword Arguments:
            step_time -- time per frame in ms (default: {80})
            step_size -- step in frames, 1 means show every frame (default: {1})
        """
        if self._animation_timer.isActive():
            self._animation_timer.stop()
        self._frame_step = step_size * self._frame_factor
        self._animation_timer.setInterval(self._time_per_frame)
        self._animation_timer.start()

    def advance_frame(self):
        firstFrame = self._frame_selector.minimum()
        lastFrame = self._frame_selector.maximum()
        new_value = self._frame_selector.value() + self._frame_step
        if new_value < firstFrame:
            self._frame_selector.setValue(firstFrame)
            self.stop_animation()
        elif new_value > lastFrame:
            self._frame_selector.setValue(lastFrame)
            self.stop_animation()
        else:
            self._frame_selector.setValue(new_value)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QLabel

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class DummyWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        source_object = kwargs.get("source_object", None)
        self._layout.addWidget(QLabel("content is missing here", self._base))
        self._configurator = {"value": "Oops!"}
        self.default_labels()
        self.update_labels()
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "The real widget that should have been here is missing! Let the authors know about it."
        self._base.setToolTip(tooltip_text)

    def configure_using_default(self):
        """This is too simple to have a default value"""

    def default_labels(self):
        """Each Widget should have a default tooltip and label,
        which will be set in this method, unless specific
        values are provided in the settings of the job that
        is being configured."""
        if self._label_text == "":
            self._label_text = "DummyWidget"
        if self._tooltip == "":
            self._tooltip = "A placeholder which tells us that a real widget is missing"

    def get_value(self):
        return self._configurator["value"]



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import os

from qtpy.QtCore import QDir
from qtpy.QtGui import QIcon


class PyQtIcons:
    def __init__(self, path):
        self.res_dir = QDir(path)
        self._icons = {}
        self.res_dir.setNameFilters(["*.png"])
        files = self.res_dir.entryList()
        for f in files:
            label = ".".join(str(f).split(".")[:-1])
            self._icons[label] = QIcon(self.res_dir.filePath(f))


ICONS = PyQtIcons(".")



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from abc import abstractmethod

import h5py
from qtpy.QtCore import QObject, Slot, Signal, QMutex, QModelIndex, Qt
from qtpy.QtGui import QStandardItemModel, QStandardItem

from MDANSE.MLogging import LOG

from MDANSE_GUI.Session.LocalSession import json_decoder


class BasicPlotDataItem(QStandardItem):
    """Each item in the PlotDataModel is a BasicPlotDataItem.
    Since there are only two levels of nesting in the data tree,
    there are two types of item we will use, both of them
    derived from BasicPlotDataItem
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._item_type = "generic"

    @abstractmethod
    def data_path(self):
        pass

    @abstractmethod
    def file_number(self):
        pass

    def populate(self, file):
        for key in file.keys():
            try:
                file[key]
            except Exception as e:
                LOG.error(f"error {e} when accessing file[{key}]")
            else:
                child = DataSetItem()
                child.setText(key)
                child.setData(key, role=Qt.ItemDataRole.DisplayRole)
                child.setData(key, role=Qt.ItemDataRole.UserRole)
                try:
                    file[key][:]
                except:
                    child._item_type = "group"
                self.appendRow(child)


class DataSetItem(BasicPlotDataItem):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._item_type = "dataset"

    def data_path(self) -> str:
        parent_path = self.parent().data_path()
        own_path = self.data(role=Qt.ItemDataRole.UserRole)
        return "/".join([parent_path, own_path])

    def file_number(self) -> int:
        return self.parent().file_number()


class DataFileItem(BasicPlotDataItem):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._item_type = "file"

    def data_path(self) -> str:
        return ""

    def file_number(self) -> int:
        return self.data(role=Qt.ItemDataRole.UserRole)


class MDADataStructure:

    def __init__(self, filename: str):
        self._file = h5py.File(filename)
        self.check_metadata()
        self.find_information()

    def close(self):
        self._file.close()

    def find_information(self):
        self._main_dataset = ""
        self._components = []
        self._axis_datasets = []
        self._supporting_datasets = []

    def check_metadata(self):
        meta_dict = {}

        def put_into_dict(name, obj):
            try:
                string = obj[:][0]
            except TypeError:
                try:
                    string = obj[0]
                except TypeError:
                    return
            try:
                string = string.decode()
            except KeyError:
                LOG.debug(f"Decode failed for {name}: {obj}")
                meta_dict[name] = str(obj)
            else:
                try:
                    meta_dict[name] = json_decoder.decode(string)
                except ValueError:
                    meta_dict[name] = string

        try:
            meta = self._file["metadata"]
        except KeyError:
            return
        else:
            meta.visititems(put_into_dict)
        self._metadata = meta_dict


class PlotDataModel(QStandardItemModel):
    """Meant to be used with DoublePanel, GeneralView
    and ItemVisualiser. It stores elements and emits
    them to the ItemVisualiser."""

    error = Signal(str)
    all_elements = Signal(object)

    def __init__(self, parent: QObject = None):
        super().__init__(parent=parent)
        self.mutex = QMutex()
        self._nodes = {}
        self._next_number = 0

    @Slot(str)
    def add_file(self, filename: str):
        try:
            new_datafile = MDADataStructure(filename)
        except Exception as e:
            LOG.error(f"Invalid: {str(e)}")
        else:
            self._nodes[self._next_number] = new_datafile
            new_item = DataFileItem()
            new_item.setData(f"{filename}", role=Qt.ItemDataRole.DisplayRole)
            new_item.setData(self._next_number, role=Qt.ItemDataRole.UserRole)
            self._next_number += 1
            self.appendRow(new_item)
            new_item.populate(new_datafile._file)

    def inner_object(self, index: QModelIndex) -> MDADataStructure:
        model_item = self.itemFromIndex(index)
        number = model_item.file_number()
        data_path = model_item.data_path()
        data_structure = self._nodes[number]
        if data_path:
            return data_structure._file[data_path]
        return data_structure

    def removeRow(self, row: int, parent: QModelIndex = None):
        self.mutex.lock()
        try:
            node_number = self.item(row).data(role=Qt.ItemDataRole.UserRole)
        except AttributeError:
            return
        data_object = self._nodes.pop(node_number)
        data_object.close()
        if parent is None:
            super().removeRow(row)
        else:
            super().removeRow(row, parent)
        self.mutex.unlock()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext

import numpy as np
import matplotlib.pyplot as mpl
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt5agg import (
    NavigationToolbar2QT as NavigationToolbar2QTAgg,
)
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QComboBox,
    QLabel,
    QGridLayout,
    QDoubleSpinBox,
)
from qtpy.QtCore import Slot, Signal, Qt

from MDANSE.MLogging import LOG

from MDANSE_GUI.Widgets.RestrictedSlider import RestrictedSlider
from MDANSE_GUI.Tabs.Plotters.Plotter import Plotter


class SliderPack(QWidget):

    new_values = Signal(object)

    def __init__(self, *args, n_sliders=2, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        layout = QGridLayout(self)
        self.setLayout(layout)
        self._labels = []
        self._sliders = []
        self._spinboxes = []
        self._current_values = []
        self._minarray = np.zeros(n_sliders)
        self._maxarray = np.ones(n_sliders)
        self._valarray = np.ones(n_sliders) * 0.5
        self._steparray = np.ones(n_sliders) * 0.01
        self._clickarray = np.array(n_sliders * [101], dtype=int)
        current_row = 0
        for n in range(n_sliders):
            label = QLabel(self)
            slider = RestrictedSlider(self)
            slider.setOrientation(Qt.Orientation.Horizontal)
            box = QDoubleSpinBox(self)
            box.setSingleStep(self._steparray[n])
            layout.addWidget(label, current_row, 0)
            layout.addWidget(slider, current_row, 1, 1, 2)
            layout.addWidget(box, current_row, 3)
            current_row += 1
            self._labels.append(label)
            self._sliders.append(slider)
            self._spinboxes.append(box)
            slider.valueChanged.connect(self.slider_to_box)
            box.valueChanged.connect(self.box_to_slider)
            box.valueChanged.connect(self.collect_values)
            self._current_values.append(0)
        slider1 = self._sliders[0]
        slider2 = self._sliders[1]
        slider1.new_limit.connect(slider2.set_lower_limit)
        slider2.new_limit.connect(slider1.set_upper_limit)

    @Slot(bool)
    def new_coupling(self, new_val: bool):
        self._sliders[0]._coupled = new_val
        self._sliders[1]._coupled = new_val

    @Slot(object)
    def new_slider_labels(self, input):
        for number, element in enumerate(input):
            self._labels[number].setText(element)

    @Slot(object)
    def new_limits(self, input):
        for number, element in enumerate(input):
            minimum, maximum, stepsize = element[0], element[1], element[2]
            clicks = int(round((maximum - minimum) / stepsize))
            self._minarray[number] = minimum
            self._maxarray[number] = maximum
            self._steparray[number] = stepsize
            self._clickarray[number] = clicks
            temp_value = self._spinboxes[number].value()
            self._sliders[number].setMaximum(clicks)
            self._spinboxes[number].setMinimum(minimum)
            self._spinboxes[number].setMaximum(maximum)
            self._spinboxes[number].setSingleStep(stepsize)
            temp_value = min(maximum, temp_value)
            temp_value = max(minimum, temp_value)
            click_value = int(round((temp_value - minimum) / stepsize))
            self._sliders[number].setValue(click_value)
            self._spinboxes[number].setValue(temp_value)

    def set_values(self, new_values: List[float]):
        nv = np.array(new_values)
        nv = np.maximum(nv, self._minarray)
        nv = np.minimum(nv, self._maxarray)
        clicks = np.round((nv - self._minarray) / self._steparray).astype(int)
        for n in range(len(nv)):
            self._spinboxes[n].setValue(nv[n])
            self._sliders[n].setValue(clicks[n])

    @Slot()
    def slider_to_box(self):
        vals = np.zeros_like(self._valarray)
        clicks = np.zeros_like(self._clickarray)
        for ns, slider in enumerate(self._sliders):
            clicks[ns] = slider.value()
        vals = self._minarray + clicks * self._steparray
        for ns, box in enumerate(self._spinboxes):
            box.setValue(vals[ns])

    @Slot()
    def box_to_slider(self):
        self.blockSignals(True)
        vals = np.zeros_like(self._valarray)
        clicks = np.zeros_like(self._clickarray)
        for ns, box in enumerate(self._spinboxes):
            vals[ns] = box.value()
        clicks = np.round((vals - self._minarray) / self._steparray).astype(int)
        for ns, slider in enumerate(self._sliders):
            slider.setValue(clicks[ns])
        self.slider_to_box()
        self.blockSignals(False)

    @Slot()
    def collect_values(self):
        result = []
        for box in self._spinboxes:
            result.append(box.value())
        self._current_values = result
        self.new_values.emit(result)


class PlotWidget(QWidget):

    change_slider_labels = Signal(object)
    change_slider_limits = Signal(object)
    reset_slider_values = Signal(bool)
    change_slider_coupling = Signal(bool)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._plotter = None
        self._sliderpack = None
        self._plotting_context = None
        self._slider_max = 100
        self.make_canvas()
        self.set_plotter("Single")

    def set_context(self, new_context: "PlottingContext"):
        self._plotting_context = new_context
        self._plotting_context._figure = self._figure

    @Slot(str)
    def set_plotter(self, plotter_option: str):
        try:
            self._plotter = Plotter.create(plotter_option)
        except:
            self._plotter = Plotter()
        self.change_slider_labels.emit(self._plotter.slider_labels())
        self.change_slider_limits.emit(self._plotter.slider_limits())
        self.change_slider_coupling.emit(self._plotter.sliders_coupled())
        self.reset_slider_values.emit(self._plotter._value_reset_needed)
        self._plotter._slider_reference = self._sliderpack
        self._sliderpack.setEnabled(False)
        self.plot_data()

    @Slot(object)
    def slider_change(self, new_values: object):
        self._plotter.handle_slider(new_values)

    @Slot(bool)
    def set_slider_values(self, reset_needed: bool):
        if reset_needed and self._sliderpack is not None:
            values = self._plotter._initial_values
            self._sliderpack.set_values(values)

    def available_plotters(self) -> List[str]:
        return [str(x) for x in Plotter.indirect_subclasses() if str(x) != "Text"]

    def plot_data(self, update_only=False):
        if self._plotter is None:
            LOG.info("No plotter present in PlotWidget.")
            return
        if self._plotting_context is None:
            return
        self._plotter.plot(
            self._plotting_context,
            self._figure,
            update_only=update_only,
            toolbar=self._toolbar,
        )

    def make_canvas(self, width=12.0, height=9.0, dpi=100):
        """Creates a matplotlib figure for plotting

        Parameters
        ----------
        width : float, optional
            Figure width in inches, by default 12.0
        height : float, optional
            Figure height in inches, by default 9.0
        dpi : int, optional
            Figure resolution in dots per inch, by default 100

        Returns
        -------
        QWidget
            a widget containing both the figure and a toolbar below
        """
        canvas = self
        layout = QVBoxLayout(canvas)
        figure = mpl.figure(figsize=[width, height], dpi=dpi, frameon=True)
        figAgg = FigureCanvasQTAgg(figure)
        figAgg.setParent(canvas)
        figAgg.updateGeometry()
        toolbar = NavigationToolbar2QTAgg(figAgg, canvas)
        toolbar.update()
        layout.addWidget(figAgg)
        slider = SliderPack(self)
        self.change_slider_labels.connect(slider.new_slider_labels)
        self.change_slider_limits.connect(slider.new_limits)
        self.change_slider_coupling.connect(slider.new_coupling)
        self.reset_slider_values.connect(self.set_slider_values)
        slider.new_values.connect(self.slider_change)
        self._sliderpack = slider
        layout.addWidget(slider)
        layout.addWidget(toolbar)
        self._figure = figure
        self._toolbar = toolbar
        plot_selector = QComboBox(self)
        layout.addWidget(plot_selector)
        plot_selector.addItems(self.available_plotters())
        plot_selector.setCurrentText("Single")
        plot_selector.currentTextChanged.connect(self.set_plotter)
        self.set_plotter(plot_selector.currentText())



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from typing import List, Optional

import numpy as np
from vtk.util.numpy_support import numpy_to_vtk
import vtk
from qtpy.QtGui import QStandardItemModel, QStandardItem, QColor
from qtpy.QtCore import Signal, Slot, QObject, Qt

from MDANSE.Chemistry.Databases import AtomsDatabase

RGB_COLOURS = []
RGB_COLOURS.append((1.00, 0.20, 1.00))  # selection
RGB_COLOURS.append((1.00, 0.90, 0.90))  # background


def ndarray_to_vtkarray(colors, scales, indices):
    """Convert the colors and scales NumPy arrays to vtk arrays.

    Args:
        colors (numpy.array): the colors
        scales (numpy.array): the scales
        n_atoms (int): the number of atoms
    """
    # define the colours
    color_scalars = numpy_to_vtk(colors)
    color_scalars.SetName("colors")

    # some scales
    scales_scalars = numpy_to_vtk(scales)
    scales_scalars.SetName("scales")

    # the original index
    index_scalars = numpy_to_vtk(indices)
    index_scalars.SetName("index")

    scalars = vtk.vtkFloatArray()
    scalars.SetNumberOfComponents(3)
    scalars.SetNumberOfTuples(scales_scalars.GetNumberOfTuples())
    scalars.CopyComponent(0, scales_scalars, 0)
    scalars.CopyComponent(1, color_scalars, 0)
    scalars.CopyComponent(2, index_scalars, 0)
    scalars.SetName("scalars")
    return scalars


class AtomEntry(QObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._index = None
        self._indices = []
        self._items = []

    def set_values(
        self, name: str, indices: List[int], colour: QColor, size: float
    ) -> List[QStandardItem]:
        self._indices = indices
        name_item = QStandardItem(str(name))
        name_item.setEditable(False)
        colour_item = QStandardItem(str(colour))
        colour_item.setData(QColor(colour), role=Qt.ItemDataRole.BackgroundRole)
        size_item = QStandardItem(str(size))
        self._items = [name_item, colour_item, size_item]
        return self._items

    def colour(self) -> QColor:
        return QColor(self._items[1].text())

    def size(self) -> float:
        return float(self._items[2].text())

    def indices(self) -> List[int]:
        return self._indices


class AtomProperties(QStandardItemModel):
    new_atom_properties = Signal(object)

    def __init__(self, *args, init_colours: list = None, **kwargs):
        super().__init__(*args, **kwargs)

        self._lut = vtk.vtkColorTransferFunction()
        if init_colours is None:
            self._colour_list = RGB_COLOURS
        else:
            self._colour_list = init_colours
        self.rebuild_colours()
        self.setHorizontalHeaderLabels(["Element", "Colour", "Radius"])
        self.itemChanged.connect(self.onNewValues)
        self._groups = []
        self._total_length = 0
        self.colours = np.array([], dtype=int)
        self.radii = np.array([], dtype=float)

    def clear_table(self):
        """This was meant to be used for cleaning up,
        but the underlying VTK object seems to survive.
        The unit test shows 5 colours in the list when there
        should be 2.
        """
        self._lut.RemoveAllPoints()

    def rebuild_colours(self):
        """Puts the colours from the internal _colour_list
        into the VTK vtkColorTransferFunction object.
        The colours stored internally in self_.colour_list
        should be the same as the colours in self._lut
        AFTER this method has been called.
        """
        self._lut.RemoveAllPoints()

        for index, colour in enumerate(self._colour_list):
            self._lut.AddRGBPoint(index, colour[0], colour[1], colour[2])

    def add_colour(self, rgb: tuple[int, int, int]) -> int:
        """Puts the colour into the list and returns the
        index of the colour in the list, to be used by VTK.

        Arguments:
            rgb -- RGB colour values for the colour

        Returns:
            int - the index of the colour in the list
        """
        float_rgb = tuple([x / 255.0 for x in rgb])
        for index, value in enumerate(self._colour_list):
            if value == float_rgb:
                return index
        new_index = len(self._colour_list)
        self._colour_list.append(float_rgb)
        self.rebuild_colours()
        return new_index

    def reinitialise_from_database(
        self,
        atoms: list[str],
        element_database: AtomsDatabase,
        dummy_size: Optional[float] = None,
    ) -> list[int]:
        """Puts colours into the list based on chemical elements list
        and a database with colour values.

        Parameters
        ----------
        atoms : list[str]
            list[str] of chemical element names
        element_database : AtomsDatabase
            a dictionary containing RGB values for chemical elements,
            typically the MDANSE_GUI.MolecularViewer.database.CHEMICAL_ELEMENTS
        dummy_size : float or None
            Override the size of the dummy atoms.

        Returns
        -------
        list[int]
            A list of indices of colours, with one numbed per atom.
        """
        self.removeRows(0, self.rowCount())
        self._groups = []
        self._total_length = 0

        all_atoms = np.array(atoms)
        unique_atoms = np.unique(all_atoms)
        indices = np.arange(len(all_atoms))
        groups = {}
        for unique in unique_atoms:
            groups[unique] = indices[np.where(all_atoms == unique)]

        colour_index_list = []
        for atom in unique_atoms:
            if (
                element_database.get_atom_property(atom, "element") == "dummy"
                and dummy_size is not None
            ):
                size = dummy_size
            else:
                size = round(element_database[atom]["vdw_radius"], 2)
            atom_entry = AtomEntry()
            rgb = [int(x) for x in element_database[atom]["color"].split(";")]
            colour_index_list.append(self.add_colour(rgb))
            item_row = atom_entry.set_values(
                atom,
                groups[atom],
                QColor(rgb[0], rgb[1], rgb[2]).name(QColor.NameFormat.HexRgb),
                size,
            )
            self.appendRow(item_row)
            self._groups.append(atom_entry)
        self._total_length = len(all_atoms)
        return colour_index_list

    @Slot()
    def onNewValues(self):
        self.rebuild_colours()
        colours = np.empty(self._total_length, dtype=int)
        radii = np.empty(self._total_length, dtype=float)
        numbers = np.arange(self._total_length)
        for entry in self._groups:
            colour = entry.colour()
            red, green, blue = colour.red(), colour.green(), colour.blue()
            vtk_colour = self.add_colour((red, green, blue))
            radius = entry.size()
            indices = entry.indices()
            radii[indices] = radius
            colours[indices] = vtk_colour
        self.radii = radii
        self.colours = colours
        self.new_atom_properties.emit((colours, radii, numbers))



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Union

from qtpy.QtWidgets import QTreeView, QAbstractItemView, QApplication, QMenu
from qtpy.QtCore import Signal, Slot, QModelIndex, Qt, QMimeData
from qtpy.QtGui import QMouseEvent, QDrag, QContextMenuEvent, QStandardItem

from MDANSE_GUI.Tabs.Visualisers.DataPlotter import DataPlotter
from MDANSE_GUI.Tabs.Visualisers.PlotDataInfo import PlotDataInfo
from MDANSE_GUI.Widgets.DataDialog import DataDialog


class PlotDataView(QTreeView):
    dataset_selected = Signal(object)
    execute_action = Signal(object)
    item_details = Signal(object)
    error = Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setHeaderHidden(True)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.click_position = None
        self.clicked.connect(self.on_select_dataset)
        # self.data_dialog = DataDialog(self)
        self._data_packet = None

    def mousePressEvent(self, e: QMouseEvent) -> None:
        self.click_position = e.position()
        if self.model() is None:
            return None
        return super().mousePressEvent(e)

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        index = self.indexAt(event.pos())
        if index.row() == -1:
            # block right click when it's not on a trajectory
            return
        model = self.model()
        qitem = model.itemFromIndex(index)
        if qitem.parent() is not None:
            model = self.model()
            item = model.itemFromIndex(index)
            text = item.text()
            mda_data_structure = model.inner_object(index)
            try:
                packet = text, mda_data_structure._file
            except AttributeError:
                packet = text, mda_data_structure.file
            self._data_packet = packet
        menu = QMenu()
        item = model.itemData(index)
        self.populateMenu(menu, item)
        menu.exec_(event.globalPos())

    def populateMenu(self, menu: QMenu, item: QStandardItem):
        for action, method in [("Delete", self.deleteNode)]:
            temp_action = menu.addAction(action)
            temp_action.triggered.connect(method)

    @Slot()
    def deleteNode(self):
        model = self.model()
        index = self.currentIndex()
        model.removeRow(index.row())
        self.item_details.emit("")

    def on_select_dataset(self, index):
        model = self.model()
        item = model.itemFromIndex(index)
        text = item.text()
        mda_data_structure = model.inner_object(index)
        try:
            packet = text, mda_data_structure._file
        except AttributeError:
            packet = text, mda_data_structure.file
        self.dataset_selected.emit(packet)
        if hasattr(mda_data_structure, "_metadata"):
            self.item_details.emit(mda_data_structure._metadata)
        else:
            try:
                text += "\n"
                for attr in mda_data_structure.attrs:
                    text += f"{attr}: {mda_data_structure.attrs[attr]}\n"
                self.item_details.emit(text)
            except:
                self.item_details.emit("No additional information included.")

    @Slot(QModelIndex)
    def item_picked(self, index: QModelIndex):
        model = self.model()
        model_item = model.itemFromIndex(index)
        item_type = model_item._item_type
        mda_data = model.inner_object(index)
        if item_type == "file":
            try:
                description = mda_data._metadata
            except AttributeError:
                description = f"File {mda_data._file.filename}, no further information"
        elif item_type == "dataset":
            dataset = model.inner_object(index)
            description = f"{dataset}{model_item.data(role=Qt.ItemDataRole.UserRole)}\n"
            for key in dataset.attrs.keys():
                description += f"{key}: {dataset.attrs[key]}\n"
        elif item_type == "group":
            dataset = model.inner_object(index)
            description = f"{dataset}{model_item.data(role=Qt.ItemDataRole.UserRole)}\n"
            for key in dataset.attrs.keys():
                description += f"{key}: {dataset.attrs[key]}\n"
        else:
            description = "generic item"
        self.item_details.emit(description)  # this should emit the job name

    def connect_to_visualiser(
        self, visualiser: Union[DataPlotter, PlotDataInfo]
    ) -> None:
        """Connect to a visualiser.

        Parameters
        ----------
        visualiser : Action or TextInfo
            A visualiser to connect to this view.
        """
        if isinstance(visualiser, DataPlotter):
            self.dataset_selected.connect(visualiser.accept_data)
        elif isinstance(visualiser, PlotDataInfo):
            self.item_details.connect(visualiser.update_panel)
        else:
            raise NotImplementedError(
                f"Unable to connect view {type(self)} to visualiser "
                f"{type(visualiser)}"
            )



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import os
import json
from typing import Dict, List

from qtpy.QtCore import QObject, Signal, Slot

from MDANSE.MLogging import LOG
from MDANSE import PLATFORM
from MDANSE_GUI.Tabs.Settings.LocalSettings import LocalSettings

json_encoder = json.encoder.JSONEncoder()
json_decoder = json.decoder.JSONDecoder()


class LocalSession(QObject):
    """Stores different parameters, and the state
    of the software.

    The intention is to have a single session that
    can be accessed by different tabs.

    At the moment, LocalSession is meant to be used
    for testing individual GUI components, and
    is largely a mock object.
    """

    new_units = Signal(dict)
    new_cmap = Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parameters = {}
        self._paths = {}
        self._units = {}
        self._colours = {}
        self._state = None
        self._filename = None
        self.populate_defaults()

    def populate_defaults(self):
        self._paths["root_directory"] = os.path.expanduser("~")
        self._units["energy"] = "meV"
        self._units["time"] = "fs"
        self._units["distance"] = "ang"
        self._units["reciprocal"] = "1/ang"
        self._colours["colormap"] = "viridis"
        self._colours["style"] = "ggplot"

    @Slot(dict)
    def update_units(self, input: Dict):
        for key, value in input.items():
            self._units[key] = value
        self.new_units.emit(self._units)

    @Slot(str)
    def update_cmap(self, input: str):
        self._colours["colormap"] = input
        self.new_cmap.emit(self._colours["colormap"])

    def get_parameter(self, key: str) -> str:
        value = self._parameters.get(key, None)
        return value

    def get_path(self, key: str) -> str:
        value = self._paths.get(key, ".")
        return value

    def set_path(self, key: str, value: str):
        self._paths[key] = value

    def get_unit(self, key: str) -> str:
        value = self._units.get(key, "1")
        return value

    def obtain_settings(self, gui_element):
        return LocalSettings()

    def sections(self) -> List[Dict[str, str]]:
        return [self._units, self._colours]

    @Slot()
    def save(self, fname: str = None):
        all_items = {}
        all_items["paths"] = self._paths
        all_items["units"] = self._units
        all_items["colours"] = self._colours
        output = json_encoder.encode(all_items)
        if fname is None:
            if self._filename is None:
                fname = os.path.join(
                    PLATFORM.application_directory(), "gui_session.json"
                )
            else:
                fname = self._filename
        try:
            with open(fname, "w") as target:
                target.write(output)
        except:
            return
        else:
            self._filename = fname

    def load(self, fname: str = None):
        if fname is None:
            fname = os.path.join(PLATFORM.application_directory(), "gui_session.json")
        try:
            with open(fname, "r") as source:
                all_items_text = source.readline()
        except:
            LOG.warning(f"Failed to read session settings from {fname}")
        else:
            all_items = json_decoder.decode(all_items_text)
            self._paths = all_items["paths"]
            self._units = all_items["units"]
            if "colours" in all_items:
                self._colours = all_items["colours"]
            self._filename = fname



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import matplotlib.pyplot as mpl
from qtpy.QtWidgets import (
    QWidget,
    QFormLayout,
    QComboBox,
    QGroupBox,
    QVBoxLayout,
)
from qtpy.QtCore import Slot, Signal

from MDANSE.MLogging import LOG
from MDANSE.Framework.Units import measure


class PlotSettings(QWidget):

    plot_settings_changed = Signal()

    def __init__(self, *args, settings=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._settings = settings
        self._unit_fields = {}
        self.plot_settings_changed.connect(self.update_settings_file)

    @Slot()
    def update_settings_file(self):
        self._settings.save_values()

    @Slot(str)
    def set_style(self, style_name: str):
        mpl_group = self._settings.group("matplotlib")
        if not mpl_group.set("style", style_name):
            mpl_group.add(
                "style",
                style_name,
                "Name of the matplotlib style to be used for plotting.",
            )
        try:
            mpl.style.use(style_name)
        except:
            LOG.error(f"Could not set matplotlib style to {style_name}")
            backup_style = self._unit_lookup._settings.default_value(
                "matplotlib", "style"
            )
            if backup_style is None:
                mpl.style.use("default")
            else:
                mpl.style.use(backup_style)
        else:
            self.plot_settings_changed.emit()

    @Slot(str)
    def set_cmap(self, cmap_name: str):
        colour_group = self._settings.group("colours")
        if not colour_group.set("colormap", cmap_name):
            colour_group.add(
                "colormap",
                cmap_name,
                "Name of the matplotlib colormap to be used in 2D plots.",
            )
        self.plot_settings_changed.emit()

    @Slot(object)
    def update_plot_details(self, input):
        """Dummy method, for compatibility only"""

    @Slot()
    def update_units(self):
        unit_group = self._settings.group("units")
        try:
            energy = self._unit_fields["energy"].currentText()
        except:
            LOG.warning("Could not get the energy unit from GUI")
        else:
            try:
                measure(1.0, "rad/ps", equivalent=True).toval(energy)
            except:
                energy = self._settings.default_value("units", "energy")
            else:
                if not unit_group.set("energy", energy):
                    unit_group.add(
                        "energy",
                        energy,
                        "Preferred physical unit for expressing energy.",
                    )
        try:
            time = self._unit_fields["time"].currentText()
        except:
            LOG.warning("Could not get the time unit from GUI")
        else:
            try:
                measure(1.0, "ps").toval(time)
            except:
                time = self._settings.default_value("units", "time")
            else:
                if not unit_group.set("time", time):
                    unit_group.add(
                        "time", time, "Preferred physical unit for expressing time."
                    )
        try:
            distance = self._unit_fields["distance"].currentText()
        except:
            LOG.warning("Could not get the distance unit from GUI")
        else:
            try:
                measure(1.0, "nm").toval(distance)
            except:
                distance = self._settings.default_value("units", "distance")
            else:
                if not unit_group.set("distance", distance):
                    unit_group.add(
                        "distance",
                        distance,
                        "Preferred physical unit for expressing distance.",
                    )
        try:
            reciprocal = self._unit_fields["reciprocal"].currentText()
        except:
            LOG.warning("Could not get the reciprocal space unit from GUI")
        else:
            try:
                measure(1.0, "1/nm").toval(reciprocal)
            except:
                reciprocal = self._settings.default_value("units", "reciprocal")
            else:
                if not unit_group.set("reciprocal", reciprocal):
                    unit_group.add(
                        "reciprocal",
                        reciprocal,
                        "Preferred physical unit for expressing (quasi)momentum, i.e. reciprocal space units.",
                    )
        self.plot_settings_changed.emit()

    def make_layout(self, width=12.0, height=9.0, dpi=100):
        """Creates a matplotlib figure for plotting

        Parameters
        ----------
        width : float, optional
            Figure width in inches, by default 12.0
        height : float, optional
            Figure height in inches, by default 9.0
        dpi : int, optional
            Figure resolution in dots per inch, by default 100

        Returns
        -------
        QWidget
            a widget containing both the figure and a toolbar below
        """
        layout = QVBoxLayout(self)
        top_layout = QFormLayout()
        style_selector = QComboBox(self)
        style_selector.addItem("default")
        style_list_mpl = mpl.style.available
        style_list_filtered = [x for x in style_list_mpl if x[0] != "_"]
        style_list_filtered = [x for x in style_list_filtered if "lorbli" not in x]
        style_list_filtered = [x for x in style_list_filtered if x not in ["fast"]]
        style_selector.addItems(style_list_filtered)
        try:
            style_string = self._settings.group("matplotlib").get("style")
        except:
            style_string = "default"
        style_selector.setCurrentText(style_string)
        style_selector.currentTextChanged.connect(self.set_style)
        top_layout.addRow("Matplotlib style:", style_selector)
        try:
            colour_group = self._settings.group("colours")
            try:
                current_cmap = colour_group.get("colormap")
            except KeyError:
                LOG.warning(f"Could not get colormap from colours")
                colour_group.add(
                    "colormap",
                    "viridis",
                    "Name of the matplotlib colormap to be used in 2D plots.",
                )
                current_cmap = "viridis"
            else:
                if current_cmap not in mpl.colormaps():
                    current_cmap = "viridis"
        except:
            LOG.warning(f"Could not get the colours group")
            current_cmap = "viridis"
        cmap_selector = QComboBox(self)
        cmap_selector.addItems(mpl.colormaps())
        cmap_selector.setCurrentText(current_cmap)
        cmap_selector.currentTextChanged.connect(self.set_cmap)
        top_layout.addRow("Colormap:", cmap_selector)
        layout.addLayout(top_layout)
        box = QGroupBox("Units", self)
        layout.addWidget(box)
        unit_layout = QFormLayout(box)
        box.setLayout(unit_layout)
        energy_combo = QComboBox(box)
        energy_combo.addItems(["meV", "1/cm", "THz"])
        energy_combo.currentTextChanged.connect(self.update_units)
        time_combo = QComboBox(box)
        time_combo.addItems(["fs", "ps", "ns"])
        time_combo.currentTextChanged.connect(self.update_units)
        distance_combo = QComboBox(box)
        distance_combo.addItems(["ang", "Bohr", "nm", "pm"])
        distance_combo.currentTextChanged.connect(self.update_units)
        reciprocal_combo = QComboBox(box)
        reciprocal_combo.addItems(["1/ang", "1/Bohr", "1/nm", "1/pm"])
        reciprocal_combo.currentTextChanged.connect(self.update_units)
        unit_layout.addRow("Energy unit:", energy_combo)
        unit_layout.addRow("Time unit:", time_combo)
        unit_layout.addRow("Distance unit:", distance_combo)
        unit_layout.addRow("Reciprocal space unit:", reciprocal_combo)
        self._unit_fields["energy"] = energy_combo
        self._unit_fields["time"] = time_combo
        self._unit_fields["distance"] = distance_combo
        self._unit_fields["reciprocal"] = reciprocal_combo
        try:
            unit_group = self._settings.group("units")
        except:
            pass
        else:
            current_energy = unit_group.get("energy")
            current_time = unit_group.get("time")
            current_distance = unit_group.get("distance")
            current_reciprocal = unit_group.get("reciprocal")
            energy_combo.setCurrentText(current_energy)
            time_combo.setCurrentText(current_time)
            distance_combo.setCurrentText(current_distance)
            reciprocal_combo.setCurrentText(current_reciprocal)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtCore import Slot
from .TextInfo import TextInfo


class JobLogInfo(TextInfo):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setStyleSheet("font-family: Courier New;")

    @Slot(object)
    def update_panel(self, incoming: object):
        msgs, levels = incoming
        text = ""
        for msg, level in zip(msgs, levels):
            if level == "WARNING":
                text += f'<span style="color:orange;">{msg}</span>\n'
            elif level == "ERROR" or level == "CRITICAL":
                text += f'<span style="color:red;">{msg}</span>\n'
            else:
                text += f"<span>{msg}</span>\n"
        filtered = self.filter(text)
        self.setHtml(filtered)



import pytest
import tempfile

from MDANSE_GUI.DataViewModel.TrajectoryHolder import FileObject

reference_bytes = b"TeStCaSeFoRtHeCaChEfuNcTiOn"
reference_hash = "7105425fa73f3e6a31b72ae2ee36235bcf1f4883b16e674b80a04d6587d995ec"


@pytest.fixture(scope="module")
def temporary_fileobject():
    fdesc, fname = tempfile.mkstemp()
    with open(fname, "wb") as target:
        target.write(reference_bytes)
    fob = FileObject()
    fob.setFilename(fname)
    return fob


def test_hash_value(temporary_fileobject):
    hash = temporary_fileobject.hash
    print(hash)
    assert hash == reference_hash



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


from qtpy.QtCore import Slot, Signal, QObject, QTimer
from qtpy.QtWidgets import QSlider


class RestrictedSlider(QSlider):

    new_limit = Signal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._lower_limit = None
        self._upper_limit = None
        self._coupled = False

    @Slot(object)
    def set_lower_limit(self, new_limit: int):
        if not self._coupled:
            return
        self._lower_limit = new_limit
        if new_limit is None:
            return
        if self.value() < new_limit:
            self.setValue(new_limit)

    @Slot(object)
    def set_upper_limit(self, new_limit: int):
        if not self._coupled:
            return
        self._upper_limit = new_limit
        if new_limit is None:
            return
        if self.value() > new_limit:
            self.setValue(new_limit)

    def setValue(self, a0: int) -> None:
        if not self._coupled:
            return super().setValue(a0)
        if self._upper_limit is not None:
            if a0 > self._upper_limit:
                return
        if self._lower_limit is not None:
            if a0 < self._lower_limit:
                return
        self.new_limit.emit(a0)
        return super().setValue(a0)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, List

from MDANSE.Framework.Units import measure
from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Plotters.Plotter import Plotter

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext


class Single(Plotter):

    def __init__(self) -> None:
        super().__init__()
        self._figure = None
        self._active_curves = []
        self._backup_curves = []
        self._backup_limits = []
        self._curve_limit_per_dataset = 12
        self.height_max, self.length_max = 0.0, 0.0

    def clear(self, figure: "Figure" = None):
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            return
        target.clear()

    def get_figure(self, figure: "Figure" = None):
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            LOG.error(f"PlottingContext can't plot to {target}")
            return
        target.clear()
        return target

    def slider_labels(self) -> List[str]:
        return ["Y offset", "X offset"]

    def slider_limits(self) -> List[str]:
        return self._number_of_sliders * [[-1.0, 1.0, 0.01]]

    def handle_slider(self, new_value: List[float]):
        super().handle_slider(new_value)
        self.offset_curves()

    def offset_curves(self):
        target = self._figure
        if target is None:
            return
        if len(self._active_curves) == 0:
            return
        new_value = self._slider_values
        saved_xmin, saved_xmax, saved_ymin, saved_ymax = self._backup_limits
        for num, curve in enumerate(self._active_curves):
            xdata = self._backup_curves[num][0]
            ydata = self._backup_curves[num][1]
            new_xdata = xdata + num * self.length_max * new_value[1]
            new_ydata = ydata + num * self.height_max * new_value[0]
            curve.set_xdata(new_xdata)
            curve.set_ydata(new_ydata)
            xmin, xmax = new_xdata.min(), new_xdata.max()
            ymin, ymax = new_ydata.min(), new_ydata.max()
            saved_xmin = min(xmin, saved_xmin)
            saved_xmax = max(xmax, saved_xmax)
            saved_ymin = min(ymin, saved_ymin)
            saved_ymax = max(ymax, saved_ymax)
        self._backup_limits = [saved_xmin, saved_xmax, saved_ymin, saved_ymax]
        self._axes[0].relim()
        self._axes[0].autoscale()
        if self._toolbar is not None:
            self._toolbar.update()
            self._toolbar.push_current()
        try:
            self._axes[0].set_xlim(saved_xmin, saved_xmax)
        except ValueError:
            LOG.error(
                f"Matplotlib could not set x limits to {saved_xmin}, {saved_xmax}"
            )
        try:
            self._axes[0].set_ylim(saved_ymin, saved_ymax)
        except ValueError:
            LOG.error(
                f"Matplotlib could not set y limits to {saved_ymin}, {saved_ymax}"
            )
        target.canvas.draw()

    def plot(
        self,
        plotting_context: "PlottingContext",
        figure: "Figure" = None,
        update_only=False,
        toolbar=None,
    ):
        self.enable_slider(False)
        target = self.get_figure(figure)
        if target is None:
            return
        if toolbar is not None:
            self._toolbar = toolbar
        self._figure = target
        xaxis_unit = None
        self._active_curves = []
        self._backup_curves = []
        axes = target.add_subplot(111)
        self._axes = [axes]
        self.apply_settings(plotting_context)
        self.height_max, self.length_max = 0.0, 0.0
        if plotting_context.set_axes() is None:
            LOG.debug("Axis check failed.")
            return
        if len(plotting_context.datasets()) == 0:
            target.clear()
            target.canvas.draw()
        for name, databundle in plotting_context.datasets().items():
            dataset, colour, linestyle, marker, _, axis_label = databundle
            try:
                best_unit, best_axis = (dataset._axes_units[axis_label], axis_label)
            except KeyError:
                best_unit, best_axis = dataset.longest_axis()
            plotlabel = dataset._labels["medium"]
            xaxis_unit = plotting_context.get_conversion_factor(best_unit)
            try:
                conversion_factor = measure(1.0, best_unit, equivalent=True).toval(
                    xaxis_unit
                )
            except:
                continue
            else:
                if dataset._n_dim == 1:
                    [temp] = axes.plot(
                        dataset._axes[best_axis] * conversion_factor,
                        dataset._data,
                        linestyle=linestyle,
                        label=plotlabel,
                        color=colour,
                    )
                    try:
                        temp.set_marker(marker)
                    except ValueError:
                        try:
                            temp.set_marker(int(marker))
                        except:
                            pass
                    self._active_curves.append(temp)
                    self._backup_curves.append([temp.get_xdata(), temp.get_ydata()])
                    self.height_max = max(self.height_max, temp.get_ydata().max())
                    self.length_max = max(self.length_max, temp.get_xdata().max())
                else:
                    multi_curves = dataset.curves_vs_axis(best_unit)
                    counter = 0
                    for key, value in multi_curves.items():
                        counter += 1
                        if counter >= self._curve_limit_per_dataset:
                            break
                        try:
                            [temp] = axes.plot(
                                dataset._axes[best_axis] * conversion_factor,
                                value,
                                label=plotlabel + ":" + dataset._curve_labels[key],
                            )
                            self._active_curves.append(temp)
                            self._backup_curves.append(
                                [temp.get_xdata(), temp.get_ydata()]
                            )
                            self.height_max = max(
                                self.height_max, temp.get_ydata().max()
                            )
                            self.length_max = max(
                                self.length_max, temp.get_xdata().max()
                            )
                        except ValueError:
                            LOG.error(
                                f"Plotting failed for {plotlabel} using {best_axis}"
                            )
                            LOG.error(f"x_axis={dataset._axes[best_axis]}")
                            LOG.error(f"values={value}")
                            return
        if len(self._backup_curves) > 1:
            self.enable_slider(True)
        if update_only:
            try:
                axes.set_xlim((self._backup_limits[0], self._backup_limits[1]))
            except ValueError:
                LOG.error(
                    f"Matplotlib could not set x limits to {self._backup_limits[0]}, {self._backup_limits[1]}"
                )
            try:
                axes.set_ylim((self._backup_limits[2], self._backup_limits[3]))
            except ValueError:
                LOG.error(
                    f"Matplotlib could not set y limits to {self._backup_limits[2]}, {self._backup_limits[3]}"
                )
        else:
            xlimits, ylimits = axes.get_xlim(), axes.get_ylim()
            self._backup_limits = [xlimits[0], xlimits[1], ylimits[0], ylimits[1]]
        if xaxis_unit is not None:
            axes.set_xlabel(xaxis_unit)
        axes.grid(True)
        axes.legend(loc=0)
        self.offset_curves()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QScrollArea,
    QLabel,
    QPushButton,
)
from qtpy.QtCore import Signal, Slot


class SinglePanel(QWidget):
    """A basic component of the GUI, it combines the
    viewer for a data model, a visualiser for a specific
    component, and a button panel for actions.
    """

    error = Signal(str)
    item_picked = Signal(object)

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        self._visualiser = None

        visualiser_side = kwargs.pop("visualiser_side", None)
        self._tab_reference = kwargs.pop("tab_reference", None)
        _ = kwargs.pop("data_side", None)

        super().__init__(*args, **kwargs)

        buffer = QWidget(self)
        scroll_area = QScrollArea()
        scroll_area.setWidget(buffer)
        scroll_area.setWidgetResizable(True)
        layout = QHBoxLayout(buffer)
        base_layout = QHBoxLayout(self)
        buffer.setLayout(layout)
        base_layout.addWidget(scroll_area)
        self.setLayout(base_layout)
        self._base = buffer

        leftside = QWidget(self._base)
        leftlayout = QVBoxLayout(leftside)
        leftside.setLayout(leftlayout)

        layout.addWidget(leftside)

        upper_buttons = QWidget(leftside)
        ub_layout = QHBoxLayout(upper_buttons)
        upper_buttons.setLayout(ub_layout)
        lower_buttons = QWidget(leftside)
        lb_layout = QHBoxLayout(lower_buttons)
        lower_buttons.setLayout(lb_layout)

        self._tab_label = QLabel(leftside)
        leftlayout.addWidget(self._tab_label)
        leftlayout.addWidget(upper_buttons)
        leftlayout.addWidget(lower_buttons)
        if visualiser_side is not None:
            self._visualiser = visualiser_side
            leftlayout.addWidget(self._visualiser)

        self._leftlayout = leftlayout
        self._lb_layout = lb_layout
        self._ub_layout = ub_layout

    def connect_logging(self):
        self.error.connect(self._tab_reference.error)
        for thing in [self._visualiser]:
            thing.error.connect(self._tab_reference.error)

    @Slot(str)
    def set_label_text(self, text: str):
        self._tab_label.setText(text)

    def add_button(self, label: str = "Button!", slot=None, upper=True):
        temp = QPushButton(label, self._base)
        if slot is not None:
            temp.clicked.connect(slot)
        if upper:
            self._ub_layout.addWidget(temp)
        else:
            self._lb_layout.addWidget(temp)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QLineEdit, QLabel
from qtpy.QtGui import QIntValidator

from MDANSE.MLogging import LOG

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class FramesWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, layout_type="QGridLayout", **kwargs)
        trajectory_configurator = kwargs.get("trajectory_configurator", None)
        if trajectory_configurator is not None:
            try:
                self._last_frame = trajectory_configurator["length"]
            except:
                self._last_frame = -1
        else:
            self._last_frame = -1
        labels = [
            QLabel("First frame", self._base),
            QLabel("Last frame", self._base),
            QLabel("in steps of", self._base),
        ]
        fields = [
            QLineEdit("0", self._base),
            QLineEdit(str(self._last_frame), self._base),
            QLineEdit("1", self._base),
        ]
        placeholders = ["0", str(self._last_frame), "1"]
        validators = [QIntValidator(parent_field) for parent_field in fields]
        for field_num in range(3):
            self._layout.addWidget(labels[field_num], 0, 2 * field_num)
            self._layout.addWidget(fields[field_num], 0, 2 * field_num + 1)
            fields[field_num].setValidator(validators[field_num])
            fields[field_num].textChanged.connect(self.updateValue)
            fields[field_num].setPlaceholderText(placeholders[field_num])
        self._fields = fields
        self._validators = validators
        self._default_values = placeholders
        self.default_labels()
        self.update_labels()
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "A single logical value that can be True of False"
        for field in fields:
            field.setToolTip(tooltip_text)

    def default_labels(self):
        """Each Widget should have a default tooltip and label,
        which will be set in this method, unless specific
        values are provided in the settings of the job that
        is being configured."""
        if self._label_text == "":
            self._label_text = "FramesWidget"
        if self._tooltip == "":
            self._tooltip = (
                "Trajectory frames to be used, given as (First, Last, StepSize)"
            )

    def value_from_configurator(self):
        if self._last_frame > 0:
            for val in self._validators:
                val.setBottom(-abs(self._last_frame))
                val.setTop(abs(self._last_frame))
        elif self._configurator.check_dependencies():
            minval, maxval = self._configurator._mini, self._configurator._maxi
            LOG.info(f"Configurator min/max: {minval}, {maxval}")
            for val in self._validators:
                val.setBottom(-abs(maxval))
                val.setTop(abs(maxval))

    def get_widget_value(self):
        result = []
        for n, field in enumerate(self._fields):
            strval = field.text()
            try:
                val = int(strval)
            except:
                val = int(self._default_values[n])
            result.append(val)
        return result



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List
import math

import numpy as np

from qtpy.QtWidgets import (
    QDialog,
    QPushButton,
    QGridLayout,
    QVBoxLayout,
    QWidget,
    QLabel,
    QComboBox,
    QLineEdit,
    QTextEdit,
    QApplication,
)
from qtpy.QtCore import Signal, Slot
from qtpy.QtGui import QDoubleValidator

import matplotlib.pyplot as mpl
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt5agg import (
    NavigationToolbar2QT as NavigationToolbar2QTAgg,
)

from MDANSE.MLogging import LOG
from MDANSE.Framework.Units import measure
from MDANSE.Framework.InstrumentResolutions.IInstrumentResolution import (
    IInstrumentResolution,
)


widget_text_map = {
    "ideal": "ideal",
    "Gaussian": "gaussian",
    "Lorentzian": "lorentzian",
    "triangular": "triangular",
    "square": "square",
    "pseudo-Voigt": "pseudovoigt",
}


gauss_denum = 2.0 * (2.0 * math.log(2.0)) ** 0.5


def convert_parameters(fwhm: float, centre: float, peak_type: str) -> List[float]:
    """Takes the values from the ResolutionDialog as input,
    and returns the sigma and mu values as expected by MDANSE.
    The conversion between sigma and FWHM is different for
    different peak functions.

    Parameters
    ----------
    fwhm : float
        The FWHM of the peak, in rad/ps
    centre : float
        The centre of the peak, in rad/ps
    peak_type : str
        Name of the function, one of the values of the widget_text_map dict

    Returns
    -------
    List[float]
        A pair of [sigma, mu] values, or empty for ideal
    """
    if peak_type == "ideal":
        return []
    elif peak_type == "triangular":
        vals = [fwhm, centre]
    elif peak_type == "square":
        vals = [fwhm / 2, centre]
    elif peak_type == "gaussian":
        vals = [fwhm / gauss_denum, centre]
    elif peak_type == "lorentzian":
        vals = [fwhm / 2, centre]
    else:
        vals = [fwhm, centre]
    return vals[0], vals[1]


def revert_parameters(values: dict, peak_type: str) -> List[float]:
    """Converts the sigma and mu values back to FWHM and peak centre.
    Used for passing the values from the main GUI back to the
    helper dialog.

    Parameters
    ----------
    values : dict
        A dictionary of all the peak parameters from the InstrumentResolutionWidget
    peak_type : str
        Name of the function, one of the values of the widget_text_map dict

    Returns
    -------
    List[float]
        A pair of [FWHM, centre] values
    """
    if peak_type == "ideal":
        return [1.0, 0]
    elif peak_type == "triangular":
        vals = [values["sigma"], values["mu"]]
    elif peak_type == "square":
        vals = [values["sigma"] * 2, values["mu"]]
    elif peak_type == "gaussian":
        vals = [values["sigma"] * gauss_denum, values["mu"]]
    elif peak_type == "lorentzian":
        vals = [values["sigma"] * 2, values["mu"]]
    elif "oigt" in peak_type:
        try:
            sigma = (
                values["sigma_gaussian"] * gauss_denum + values["sigma_lorentzian"] * 2
            )
        except KeyError:
            sigma = 2.0
        try:
            mu = values["mu_gaussian"] + values["mu_lorentzian"]
        except KeyError:
            mu = 0.0
        vals = [sigma / 2, mu / 2]
    else:
        vals = [values["sigma"], values["mu"]]
    return vals[0], vals[1]


class ResolutionCalculator:

    def __init__(self) -> None:
        self._valid = False
        self._fwhm_value = 1.0
        self._centre_value = 0.0
        self._unit_value = "meV"
        self._eta_value = 0.0

    def update_model(self, new_model: str):
        """Activated on new value in the peak type combo box,
        creates the new IInstrumentResolution subclass.

        Parameters
        ----------
        new_model : str
            Name of the resolution class to create.
        """
        self._resolution_name = new_model
        self._resolution = IInstrumentResolution.create(widget_text_map[new_model])
        self._resolution.build_configuration()
        self.recalculate_peak(
            self._fwhm_value, self._centre_value, self._eta_value, self._unit_value
        )

    def recalculate_peak(self, fwhm, centre, eta, unit):
        """The main method of the dialog.
        Collects the inputs from text fields and
        calculates the peak function on an array
        of points around the peak centre.
        """
        factor = measure(1.0, iunit=unit, equivalent=True).toval("rad/ps")
        self._factor_value = factor
        self._fwhm_value = float(fwhm)
        self._centre_value = float(centre)
        self._unit_value = unit
        try:
            self._eta_value = float(eta)
        except TypeError:
            self._eta_value = None
        if "oigt" in self._resolution_name:
            gauss_sigma, gauss_mu = convert_parameters(
                self._fwhm_value, self._centre_value, "gaussian"
            )
            lorentz_sigma, lorentz_mu = convert_parameters(
                self._fwhm_value, self._centre_value, "lorentzian"
            )
            self.set_peak_parameter(lorentz_mu * factor, "mu_lorentzian")
            self.set_peak_parameter(gauss_mu * factor, "mu_gaussian")
            self.set_peak_parameter(lorentz_sigma * factor, "sigma_lorentzian")
            self.set_peak_parameter(gauss_sigma * factor, "sigma_gaussian")
            self.set_peak_parameter(self._eta_value, "eta")
        else:
            temp_name = widget_text_map[self._resolution_name]
            try:
                sigma, mu = convert_parameters(
                    self._fwhm_value, self._centre_value, temp_name
                )
            except ValueError:
                self._fwhm_value = 0.0
            else:
                self.set_peak_parameter(mu * factor, "mu")
                self.set_peak_parameter(sigma * factor, "sigma")
        extra_width = abs(self._fwhm_value)
        if extra_width <= 1e-14:
            extra_width = 1
        self._omega_axis = np.linspace(
            factor * (centre - 3 * extra_width),
            factor * (centre + 3 * extra_width),
            501,  # odd number is needed for 'ideal' function to work
        )
        try:
            self._resolution.set_kernel(self._omega_axis, 1.0)
        except:
            self._valid = False
        else:
            self._valid = True

    def set_peak_parameter(self, value: float, key: str):
        """A convenience method for passing an input value
        to the IInstrumentResolution class instance.

        Parameters
        ----------
        value : float
            numerical value of the peak parameter
        key : str
            name of the parameter from the 'settings' dictionary
        """
        try:
            self._resolution._configuration[key].configure(value)
        except KeyError:
            LOG.warning(f"Could not find {key} in {self._resolution._configuration}")

    def summarise_results(self, rounding_precision=3):
        results = {"function": self._resolution_name}
        text = ""
        for key, value in self._resolution._configuration.items():
            original_value = value["value"]
            if abs(original_value) < 1e-12:
                rounded_value = 0.0
            else:
                rounded_value = round(
                    original_value,
                    abs(math.floor(math.log10(abs(original_value))))
                    + rounding_precision,
                )
            text += f"settings[{key}] = {rounded_value}\n"
            results[key] = rounded_value
        return text, results


class ResolutionWidget(QWidget):

    parameters_changed = Signal(dict)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._calculator = ResolutionCalculator()

        layout = QGridLayout(self)
        self.setWindowTitle("Resolution Helper for MDANSE")
        self._resolution = None
        self.setLayout(layout)
        self._peak_selector = QComboBox(self)
        self._peak_selector.addItems(widget_text_map.keys())
        self._unit_selector = QComboBox(self)
        self._unit_selector.addItems(["meV", "1/cm", "THz"])
        self._fwhm = QLineEdit("1.0", self)
        self._centre = QLineEdit("0.0", self)
        self._eta = QLineEdit("0.0", self)
        self._eta.setEnabled(False)
        for ledit in [self._fwhm, self._centre, self._eta]:
            ledit.setValidator(QDoubleValidator(ledit))
            ledit.setPlaceholderText("N/A")
        self._output_field = QTextEdit(self)
        self._output_field.setReadOnly(True)
        self._apply_button = QPushButton("Apply", self)
        self._apply_button.clicked.connect(self.apply_changes)
        text_labels = ["Peak function", "Energy unit", "FWHM", "Centre", "eta"]
        for number, widget in enumerate(
            [
                self._peak_selector,
                self._unit_selector,
                self._fwhm,
                self._centre,
                self._eta,
            ]
        ):
            layout.addWidget(QLabel(text_labels[number], self), number, 0)
            layout.addWidget(widget, number, 1)
        layout.addWidget(self._output_field, number + 1, 0, 1, 2)
        layout.addWidget(self._apply_button, number + 2, 0)
        canvas = self.make_canvas()
        layout.addWidget(canvas, 0, 2, 8, 1)
        for widget in [self._fwhm, self._centre, self._eta]:
            widget.textChanged.connect(self.recalculate_peak)
        self._peak_selector.currentTextChanged.connect(self.update_model)
        self._unit_selector.currentTextChanged.connect(self.recalculate_peak)
        self.update_model(self._peak_selector.currentText())

    def make_canvas(self, width=12.0, height=9.0, dpi=100):
        """Creates a matplotlib figure for plotting

        Parameters
        ----------
        width : float, optional
            Figure width in inches, by default 12.0
        height : float, optional
            Figure height in inches, by default 9.0
        dpi : int, optional
            Figure resolution in dots per inch, by default 100

        Returns
        -------
        QWidget
            a widget containing both the figure and a toolbar below
        """
        canvas = QWidget(self)
        layout = QVBoxLayout(canvas)
        figure = mpl.figure(figsize=[width, height], dpi=dpi, frameon=True)
        figAgg = FigureCanvasQTAgg(figure)
        figAgg.setParent(canvas)
        figAgg.updateGeometry()
        toolbar = NavigationToolbar2QTAgg(figAgg, canvas)
        toolbar.update()
        layout.addWidget(figAgg)
        layout.addWidget(toolbar)
        self._figure = figure
        return canvas

    @Slot(str)
    def update_model(self, new_model: str):
        """Activated on new value in the peak type combo box,
        creates the new IInstrumentResolution subclass.

        Parameters
        ----------
        new_model : str
            Name of the resolution class to create.
        """
        self._resolution_name = new_model
        self._calculator.update_model(new_model)
        if "oigt" in new_model:
            for field in [self._fwhm, self._centre, self._eta]:
                field.setEnabled(True)
        elif "deal" in new_model:
            for field in [self._fwhm, self._centre, self._eta]:
                field.setEnabled(False)
        else:
            for field in [self._fwhm, self._centre]:
                field.setEnabled(True)
            self._eta.setEnabled(False)
        self.recalculate_peak()

    @Slot()
    def recalculate_peak(self):
        """The main method of the dialog.
        Collects the inputs from text fields and
        calculates the peak function on an array
        of points around the peak centre.
        """
        try:
            fwhm = abs(float(self._fwhm.text()))
        except:
            return
        try:
            centre = float(self._centre.text())
        except:
            return
        if "oigt" in self._resolution_name:
            try:
                eta = float(self._eta.text())
            except:
                return
        else:
            eta = None
        unit = self._unit_selector.currentText()
        self._calculator.recalculate_peak(fwhm, centre, eta, unit)
        if self._calculator._valid:
            self._apply_button.setEnabled(True)
            self.update_text_output()
            self.update_plot()
        else:
            self.update_text_output(error=True)
            self._apply_button.setEnabled(False)

    def update_fields(self, widget_values):
        """Method for passing the values from the main
        GUI into the helper dialog. Converts all the
        numbers from MDANSE inputs (sigma in rad/ps)
        to the ResolutionDialog values (FWHM in users's preferred units).
        During the execution of this method the signals
        of the widgets are blocked, not to trigger
        the sending of the input values which would
        overwrite the GUI values.

        Parameters
        ----------
        widget_values : tuple[str, dict]
            Widget values from InstrumentResolutionWidget
        """
        self.blockSignals(True)
        new_function_name = widget_values[0]
        offical_name = "missing"
        for key, value in widget_text_map.items():
            if new_function_name == value:
                offical_name = key
        new_params = widget_values[1]
        new_eta = new_params.get("eta", "0.0")
        try:
            fwhm, centre = revert_parameters(new_params, new_function_name)
        except:
            self.blockSignals(False)
            return
        self._peak_selector.setCurrentText(offical_name)
        if abs(fwhm) < 1e-12:
            new_fwhm = 0.0
        else:
            temp_value = fwhm / self._calculator._factor_value
            new_fwhm = round(
                temp_value,
                abs(math.floor(math.log10(abs(temp_value)))) + 3,
            )
        if abs(centre) < 1e-12:
            new_centre = 0.0
        else:
            temp_value = centre / self._calculator._factor_value
            new_centre = round(
                temp_value,
                abs(math.floor(math.log10(abs(temp_value)))) + 3,
            )
        self._fwhm.setText(str(new_fwhm))
        self._centre.setText(str(new_centre))
        self._eta.setText(str(new_eta))
        self.blockSignals(False)
        self.recalculate_peak()

    def update_text_output(self, rounding_precision=3, error=False):
        """Updates the text in the QTextEdit widget.
        It shows the user what MDANSE values will be
        created out of the current inputs.
        Rounding is applied to the numbers for legibility.

        Parameters
        ----------
        rounding_precision : int, optional
            number of significant places to include, by default 3
        """
        if error:
            text = "Current parameters do not produce a valid peak function.\n"
            text += "If your FWHM=0, you may need to change it."
            self._output_field.setText(text)
            return
        text = "Parameters in MDANSE internal units\n"
        temp_text, results = self._calculator.summarise_results(
            rounding_precision=rounding_precision
        )
        self._output_field.setText(text + temp_text)
        self._results = results

    @Slot()
    def apply_changes(self):
        self.parameters_changed.emit(self._results)

    def update_plot(self):
        """Plots the latest peak function in the
        matplotlib figure.
        The plot has two x axes: one in user's units,
        and the other in MDANSE units.
        A dashed line indicates the full width at
        half maximum of the peak.
        """
        self._figure.clear()
        axes = self._figure.add_axes(111)
        axes.plot(
            self._calculator._omega_axis, self._calculator._resolution._omegaWindow
        )
        hh = np.max(self._calculator._resolution._omegaWindow) / 2
        xs = (
            np.array(
                [
                    -self._calculator._fwhm_value / 2,
                    -self._calculator._fwhm_value / 2,
                    self._calculator._fwhm_value / 2,
                    self._calculator._fwhm_value / 2,
                ]
            )
            + self._calculator._centre_value
        )
        ys = np.array([0.0, hh, hh, 0.0])
        if not "deal" in self._resolution_name:
            axes.plot(xs * self._calculator._factor_value, ys, "r:")
        axes.grid(True)
        scale = self._calculator._factor_value
        second_axis = axes.secondary_xaxis(
            "top", functions=(lambda x: x / scale, lambda x: x / scale)
        )
        axes.set_xlabel("Energy [rad/ps]")
        second_axis.set_xlabel(f"Energy [{self._calculator._unit_value}]")
        self._figure.canvas.draw()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QLineEdit, QLabel, QCheckBox
from qtpy.QtGui import QDoubleValidator
from qtpy.QtCore import Slot, Qt

from MDANSE.MLogging import LOG
from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class UnitCellWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, layout_type="grid", **kwargs)
        self._changing_label = QLabel("Unit cell", parent=self._base)
        self._unit_label = QLabel("(nm)", parent=self._base)
        self._apply_box = QCheckBox("Apply unit cell", self._base)
        self._apply_box.setTristate(False)
        self._apply_box.checkStateChanged.connect(self.toggle_widgets)
        self._apply_box.setChecked(False)
        self._layout.addWidget(self._changing_label, 0, 0)
        self._layout.addWidget(self._unit_label, 1, 0)
        self._layout.addWidget(self._apply_box, 2, 0)
        self._array_fields = {}
        self._configurator.update_trajectory_information()
        try:
            start_values = self._configurator._recommended_cell
        except AttributeError:
            start_values = self._configurator._default
        for row in range(3):
            for column in range(3):
                temp = QLineEdit(str(round(start_values[row][column], 5)), self._base)
                temp.setValidator(QDoubleValidator())
                temp.setEnabled(False)
                self._layout.addWidget(temp, row, column + 1)
                self._array_fields[(row, column)] = temp
        self._mode = 0
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = (
                "This unit cell will be set for every frame of the output trajectory"
            )
        for wid in self._array_fields.values():
            wid.setToolTip(tooltip_text)

    @Slot()
    def toggle_widgets(self):
        if self._apply_box.checkState() == Qt.CheckState.Checked:
            for widget in self._array_fields.values():
                widget.setEnabled(True)
        else:
            for widget in self._array_fields.values():
                widget.setEnabled(False)

    def configure_using_default(self):
        """This is too complex to have a default value"""

    def get_widget_value(self):
        """Collect the results from the input widgets and return the value."""
        array = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        apply = self._apply_box.checkState() == Qt.CheckState.Checked
        if apply:
            for key, value in self._array_fields.items():
                try:
                    array[key[0]][key[1]] = float(value.text())
                except:
                    LOG.error(
                        f"Could not set value ({key[0]}, {key[1]}) to {value.text()}"
                    )
        return array, apply



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import os
from typing import Dict

from qtpy.QtCore import QObject, Signal, Slot, Qt, QModelIndex
from qtpy.QtGui import QStandardItem, QStandardItemModel
import tomlkit
from tomlkit.parser import ParseError
from tomlkit.toml_file import TOMLFile

from MDANSE import PLATFORM
from MDANSE.MLogging import LOG
from MDANSE.Framework.Units import measure, unit_lookup


class UserSettingsModel(QStandardItemModel):

    file_loaded = Signal(str)

    def __init__(self, *args, settings_filename: str = "", **kwargs):
        super().__init__(*args, **kwargs)
        if settings_filename:
            self._settings = SettingsFile(settings_filename)
            self._settings.load_from_file()
        else:
            LOG.warning(f"Called UserSettingsModel without settings_filename")
            return
        self._entries_present = {}
        self._groups_present = {}
        self.populate_model()
        self.setHeaderData(0, Qt.Orientation.Horizontal, "Item")
        self.setHeaderData(1, Qt.Orientation.Horizontal, "Value")
        self.setHeaderData(2, Qt.Orientation.Horizontal, "Comment")

        self.dataChanged.connect(self.save_new_value)
        # self.scan_model()

    def refresh(self):
        self._settings.load_from_file()
        self.populate_model()
        return

    @Slot()
    def append_group(self):
        self.appendRow(
            [
                QStandardItem("new group, please rename"),
                QStandardItem(),
                QStandardItem("# new section"),
            ]
        )

    @Slot(int)
    def append_child(self, parent_row_number: int):
        parent_item = self.item(parent_row_number, 0)
        parent_item.appendRow(
            [
                QStandardItem("item"),
                QStandardItem("dummy_value"),
                QStandardItem("# new item!"),
            ]
        )

    def populate_model(self):
        for number, groupname in enumerate(self._settings.keys()):
            group = self._settings.group(groupname)
            section = group._name
            section_item = QStandardItem(section)
            section_item.setData(section)
            section_item.setEditable(False)
            section_comment = group._group_comment
            section_comment_item = QStandardItem(section_comment)
            if section not in self._groups_present:
                self.appendRow([section_item, QStandardItem(), section_comment_item])
                self._groups_present[section] = section_comment_item.index()
            for key, value in group.as_dict().items():
                key_item, value_item = QStandardItem(key), QStandardItem(value)
                key_item.setData(key)
                value_item.setData(value)
                comment = group._comments[key]
                comment_item = QStandardItem(comment)
                comment_item.setData(comment)
                key_item.setEditable(False)
                if (section, key) not in self._entries_present:
                    section_item.appendRow([key_item, value_item, comment_item])
                    self._entries_present[(section, key)] = [
                        value_item.index(),
                        comment_item.index(),
                    ]

    def scan_model(self):
        """This is meant to be used for debugging purposes only"""

        def scan_children(parent_item: QStandardItem):
            for row in range(parent_item.rowCount()):
                for column in range(parent_item.columnCount()):
                    item = parent_item.child(row, column)
                    LOG.debug(f"row={row}, column={column}, data={item.data()}")
                    if item.hasChildren():
                        scan_children(item)

        scan_children(self.invisibleRootItem())

    @Slot(QModelIndex)
    def save_new_value(self, item_index: QModelIndex):
        item = self.itemFromIndex(item_index)
        item_key = None
        row_number = item.row()
        column_number = item.column()
        new_value = item.data(role=Qt.ItemDataRole.DisplayRole)
        if item.parent() is not None:
            item_key = item.parent().child(row_number, 0).data()
            group_key = item.parent().data()
        else:
            group_key = new_value
        try:
            group = self._settings.group(group_key)
            if item_key is not None:
                if column_number == 1:
                    group.set(item_key, new_value)
                elif column_number == 2:
                    group.set_comment(item_key, new_value)
        except:
            LOG.warning(
                f"Could not store {new_value} in group[{group_key}]->[{item_key}]"
            )
            LOG.debug(group.as_toml())

    def modify_group(self, group_name: str, new_value: str, column_number: int):
        if column_number == 2:
            group = self._settings.group(group_name)
            group.set_group_comment(new_value)

    def modify_item(
        self, group_name: str, item_name: str, new_value: str, column_number: int
    ):
        group = self._settings.group(group_name)
        group._group_comment = new_value
        if column_number == 1:
            if not group.set(item_name, new_value):
                LOG.warning(
                    f"Modify item: could not set item {item_name} to value {new_value} in group {group_name}"
                )
                LOG.debug(group.as_toml())
        elif column_number == 2:
            if not group.set_comment(item_name, new_value):
                LOG.warning(
                    f"Modify item: could not set comment {item_name} to value {new_value} in group {group_name}"
                )
                LOG.debug(group.as_toml())

    @Slot("QStandardItem*")
    def on_value_changed(self, item: "QStandardItem"):
        index = item.index()
        column = index.column()
        row = index.row()
        if column < 1 or column > 2:
            return
        new_contents = index.data(role=Qt.ItemDataRole.DisplayRole)
        group_item = item.parent()
        if group_item is None:
            self.modify_group(item_key, new_contents, column)
        else:
            item_key = group_item.child(row, 0).data(role=Qt.ItemDataRole.DisplayRole)
            group_key = group_item.data(role=Qt.ItemDataRole.DisplayRole)
            self.modify_item(group_key, item_key, new_contents, column)

    @Slot()
    def writeout_settings(self):
        self._settings.save_values()


class SettingsGroup:

    def __init__(self, group_name: str) -> None:
        self._name = group_name
        self._settings = {}
        self._comments = {}
        self._group_comment = ""

    def set_group_comment(self, comment: str):
        self._group_comment = comment

    def add(self, varname: str, value: str, comment: str):
        if varname not in self._settings:
            self._settings[varname] = value
            self._comments[varname] = comment

    def set(self, varname: str, value: str):
        if not varname in self._settings:
            LOG.warning(
                f"Group {self._name} has no entry {varname}. Add it first using add()."
            )
            return False
        self._settings[varname] = value
        return True

    def set_comment(self, varname: str, value: str):
        if not varname in self._settings:
            LOG.warning(
                f"Group {self._name} has no entry {varname}. Add it first using add()."
            )
            return False
        self._comments[varname] = value
        return True

    def get(self, varname: str):
        return self._settings[varname]

    def get_comment(self, varname: str):
        return self._comments[varname]

    def populate(self, settings: Dict, comments: Dict):
        for key, value in settings.items():
            self._settings[key] = value
            self._comments[key] = comments.get(key, "---")

    def update(self, settings: Dict, comments: Dict):
        for key, value in settings.items():
            if key in self._settings:
                continue
            self._settings[key] = value
            self._comments[key] = comments.get(key, "---")

    def compare(self, settings: Dict, comments: Dict):
        obsolete_values, obsolete_comments = [], []
        for key in self._settings:
            if key not in settings:
                obsolete_values.append(key)
        for key in self._comments:
            if key not in comments:
                obsolete_comments.append(key)
        return obsolete_values, obsolete_comments

    def as_toml(self):
        results = tomlkit.table()
        results.comment(self._group_comment)
        for key in self._settings.keys():
            results[key] = self._settings[key]
            results[key].comment(self._comments.get(key, "---"))
        return results

    def as_dict(self):
        return self._settings


class SettingsFile:

    def __init__(self, name, settings_path: str = None):

        if settings_path is None:
            settings_path = PLATFORM.application_directory()
        self._top_name = name
        self._filename = os.path.join(settings_path, name + ".toml")
        self._tomldoc = None
        self._file = TOMLFile(self._filename)
        self._groups = {}
        self._defaults = {}
        self.load_from_file()

    def load_from_file(self) -> bool:
        file = TOMLFile(self._filename)
        try:
            self._tomldoc = file.read()
        except FileNotFoundError:
            LOG.warning(f"File {self._filename} does not exists.")
            return False
        except ParseError:
            LOG.warning(f"File {self._filename} could not be parsed.")
            return False
        else:
            for key in self._tomldoc.keys():
                table = self._tomldoc[key]
                group = self._groups.get(key, SettingsGroup(key))
                temp_values, temp_comments = {}, {}
                for inner_key in table.keys():
                    temp_values[inner_key] = table[inner_key]
                    temp_comments[inner_key] = table[inner_key].trivia.comment
                group.populate(temp_values, temp_comments)
                self._groups[key] = group
            return True

    def keys(self):
        return self._groups.keys()

    @property
    def groups(self):
        return self._groups

    def default_value(self, group: str, key: str):
        try:
            result = self._defaults[(group, key)]
        except KeyError:
            result = None
        LOG.warning(
            f"Bad entry in {self._filename} for {group}->{key}? Default value {result} requested."
        )
        return result

    def group(self, group_name: str) -> "SettingsGroup":
        try:
            group = self._groups[group_name]
        except KeyError:
            group = SettingsGroup(group_name)
            self._groups[group_name] = group
        return group

    def save_values(self):
        newdoc = tomlkit.document()
        for gr in self._groups.values():
            newdoc[gr._name] = gr.as_toml()
        file = TOMLFile(self._filename)
        file.write(newdoc)

    def overwrite_settings(self, group_name, values, comments):
        group = self.group(group_name)
        group.populate(values, comments)

    def extend_settings(self, group_name, values, comments):
        group = self.group(group_name)
        group.update(values, comments)
        for key, value in values.items():
            self._defaults[(group_name, key)] = value

    def check_settings(self, group_name, values, comments):
        group = self.group(group_name)
        unused_values, unused_comments = group.compare(values, comments)
        for val in unused_values:
            LOG.warning(
                f"Unnecessary entry {val} in file {self._filename}, group {group._name}"
            )


class StructuredSession(QObject):
    """Stores settings of different parts of the GUI."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._models = {}
        self._configs = {}
        self._state = None
        self._main_config_name = "mdanse_general_settings"
        self._filename = kwargs.get("filename", self._main_config_name)
        self.populate_defaults()

    def connect_settings_file(self, name: str):
        model = UserSettingsModel(settings_filename=name)
        sf = model._settings
        self._configs[name] = sf
        self._models[name] = model
        return sf

    @Slot()
    def save(self):
        for _, config in self._configs.items():
            config.save_values()

    def load(self, fname: str = None):
        """Included for compatibility with LocalSession only.
        Now each component loads its own config separately."""

    def main_settings(self):
        return self._configs[self._main_config_name]

    def obtain_settings(self, gui_element):
        try:
            name = gui_element._name
        except AttributeError:
            name = self._main_config_name
        try:
            sf = self._configs[name]
        except KeyError:
            sf = self.connect_settings_file(name)
        sf.load_from_file()
        try:
            setting_groups = gui_element.grouped_settings()
        except AttributeError:
            LOG.debug(
                f"GUI element {gui_element} did not have a grouped_settings method."
            )
        else:
            LOG.debug(f"Initialising values in {setting_groups[0]}")
            for group in setting_groups:
                gname, settings, comments = group[0], group[1], group[2]
                sf.extend_settings(gname, settings, comments)
                sf.check_settings(gname, settings, comments)
            sf.save_values()
        return sf

    def settings_model(self, settings_filename: str = ""):
        if settings_filename not in self._models:
            LOG.debug(f"session connecting to {settings_filename} for the first time")
            self.connect_settings_file(settings_filename)
        model = self._models[settings_filename]
        LOG.debug(f"session re-using the model {settings_filename}")
        self._configs[settings_filename].load_from_file()
        model.refresh()
        return model

    def populate_defaults(self):
        gs = self.connect_settings_file(self._main_config_name)
        gs.load_from_file()
        paths = gs.group("paths")
        paths.set_group_comment("Lookup of working directory paths for the main GUI")
        units = gs.group("units")
        units.set_group_comment(
            "The GUI will, where possible and indicated, use these physical units."
        )
        paths.add(
            "root_directory",
            os.path.expanduser("~"),
            "Starting path for any file search",
        )
        units.add("energy", "meV", "The unit of energy preferred by the user.")
        units.add("time", "fs", "The unit of time preferred by the user.")
        units.add("distance", "ang", "The unit of distance preferred by the user")
        units.add(
            "reciprocal", "1/ang", "The momentum (transfer) unit preferred by the user"
        )
        gs.save_values()
        self._configs[self._main_config_name] = gs

    def get_path(self, key: str) -> str:
        settings = self._configs[self._main_config_name]
        group = settings["paths"]
        value = group.get(key, ".")
        return value

    def set_path(self, key: str, value: str):
        settings = self._configs[self._main_config_name]
        group = settings["paths"]
        group.set(key, value)

    def get_unit(self, key: str) -> str:
        settings = self._configs[self._main_config_name]
        group = settings["units"]
        value = group.get(key, "N/A")
        return value



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, List

import numpy as np
from qtpy.QtWidgets import (
    QDialog,
    QPushButton,
    QVBoxLayout,
    QHBoxLayout,
    QComboBox,
    QTextEdit,
)
from qtpy.QtCore import Signal, Slot, QObject

from MDANSE.Framework.Jobs.IJob import IJob


class DataDialog(QDialog):
    new_style = Signal(str)
    icon_swap = Signal(bool)

    def __init__(
        self,
        *args,
        input_data: List[np.ndarray],
        input_units: List[str],
        is_input=False,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self._data = input_data

        layout = QVBoxLayout(self)
        self.setLayout(layout)
        top_layout = QHBoxLayout()
        layout.addChildLayout(top_layout)
        self._input_units = input_units
        self._unit_boxes = [QComboBox(self), QComboBox(self), QComboBox(self)]
        for box in self._unit_boxes:
            top_layout.addWidget(box)
        self._display = QTextEdit(self)
        layout.addWidget(self._display)
        if is_input:
            self._confirm = QPushButton("Apply", self)
            self._confirm.clicked.connect(self.commitChanges)
            layout.addWidget(self._confirm)

    def commitChanges(self):
        text = self._display.document().toPlainText()
        self.new_style.emit(text)
        label = self._selector.currentText()
        if "dark" in label:
            self.icon_swap.emit(True)
        else:
            self.icon_swap.emit(False)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtCore import QObject, Slot, Signal, QMutex, QModelIndex, Qt
from qtpy.QtGui import QStandardItemModel, QStandardItem


class GeneralModel(QStandardItemModel):
    """Meant to be used with DoublePanel, GeneralView
    and ItemVisualiser. It stores elements and emits
    them to the ItemVisualiser."""

    error = Signal(str)
    all_elements = Signal(object)

    def __init__(self, parent: QObject = None):
        super().__init__(parent=parent)
        self.mutex = QMutex()
        self._node_numbers = []
        self._nodes = {}
        self._next_number = 0

    @Slot(tuple)
    def append_object(self, input: tuple):
        thing, label = input
        self.mutex.lock()
        self._nodes[self._next_number] = thing
        self._node_numbers.append(self._next_number)
        retval = int(self._next_number)
        self._next_number += 1
        item = QStandardItem(label)
        item.setData(retval)
        self.appendRow(item)
        self.mutex.unlock()
        self.summarise_items()
        return retval

    @Slot(tuple)
    def append_object_and_embed(self, input: tuple):
        thing, label = input
        self.mutex.lock()
        self._nodes[self._next_number] = thing
        self._node_numbers.append(self._next_number)
        retval = int(self._next_number)
        self._next_number += 1
        item = QStandardItem(label)
        item.setData(retval)
        thing._list_item = item
        thing._model_index = retval
        self.appendRow(item)
        self.mutex.unlock()
        self.summarise_items()
        return retval

    def summarise_items(self):
        result = []
        self.mutex.lock()
        for nrow in range(self.rowCount()):
            index = self.index(nrow, 0)
            item = self.itemFromIndex(index)
            result.append([item.text(), item.data()])
        self.mutex.unlock()
        self.all_elements.emit(result)

    def removeRow(self, row: int, parent: QModelIndex = None):
        self.mutex.lock()
        try:
            node_number = self.item(row).data()
        except AttributeError:
            return
        self._nodes.pop(node_number)
        self._node_numbers.pop(self._node_numbers.index(node_number))
        if parent is None:
            super().removeRow(row)
        else:
            super().removeRow(row, parent)
        self.mutex.unlock()
        self.summarise_items()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from qtpy.QtWidgets import QComboBox


def highlight_default_value(field: "QComboBox") -> None:
    """Takes the input QComboBox widget and changes the font of the currently
    selected item to bold typeface. This is meant to indicate to the users
    which option was originally selected.

    Parameters
    ----------
    field : QComboBox
        The changes will be applied to the data model of this input widget
    """
    model = field.model()
    row = field.currentIndex()
    item = model.item(row, 0)
    newfont = item.font()
    newfont.setBold(True)
    item.setFont(newfont)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QVBoxLayout, QTabWidget, QTabBar
from qtpy.QtCore import Slot, Signal

from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Visualisers.PlotWidget import PlotWidget
from MDANSE_GUI.Tabs.Visualisers.DataWidget import DataWidget
from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext


class PlotHolder(QTabWidget):
    """This is the top-layer widget which will host any number of
    PlotWidgets.
    """

    error = Signal(str)

    def __init__(self, *args, unit_lookup=None, **kwargs):
        super().__init__(*args, **kwargs)

        self._settings = None
        self._last_number = 1
        layout = QVBoxLayout(self)
        self._context = []
        self._plotter = []
        self._unit_lookup = unit_lookup
        self._current_id = -1
        self.setLayout(layout)
        self._current_id = self.new_plot("Preview")
        self._protected_id = int(self._current_id)
        self.setTabsClosable(True)
        self.tabCloseRequested.connect(self.clean_up_closed_tab)
        # remove the close button on the protected tab
        tabbar = self.tabBar()
        close_button = tabbar.tabButton(
            self._protected_id, QTabBar.ButtonPosition.RightSide
        )
        close_button.deleteLater()
        tabbar.setTabButton(self._protected_id, QTabBar.ButtonPosition.RightSide, None)

    @Slot(str)
    def new_plot(self, tab_name: str) -> int:
        if not tab_name:
            tab_name = f"New plot {self._last_number}"
            self._last_number += 1
        plotting_context = PlottingContext(unit_lookup=self._unit_lookup)
        plotting_context.needs_an_update.connect(self.update_plots)
        plotter = PlotWidget(self)
        plotter.set_context(plotting_context)
        tab_id = self.addTab(plotter, tab_name)
        LOG.info(f"PlotHolder created tab: {tab_id}")
        self._context.append(plotting_context)
        self._plotter.append(plotter)
        self.setCurrentIndex(tab_id)
        return tab_id

    @Slot(str)
    def new_text(self, ignored_name: str) -> int:
        tab_name = f"New text view {self._last_number}"
        self._last_number += 1
        plotting_context = PlottingContext(unit_lookup=self._unit_lookup)
        plotting_context.needs_an_update.connect(self.update_plots)
        plotter = DataWidget(self)
        plotter.set_context(plotting_context)
        tab_id = self.addTab(plotter, tab_name)
        LOG.info(f"PlotHolder created tab: {tab_id}")
        self._context.append(plotting_context)
        self._plotter.append(plotter)
        self.setCurrentIndex(tab_id)
        return tab_id

    @Slot(int)
    def clean_up_closed_tab(self, tab_id: int):
        if tab_id == self._protected_id:
            return
        valid_id_values = [int(idnum) for idnum in range(len(self._plotter))]
        if tab_id in valid_id_values:
            valid_id_values.pop(valid_id_values.index(tab_id))
        if tab_id < len(self._context):
            self._context.pop(tab_id)
        if tab_id < len(self._plotter):
            self._plotter.pop(tab_id)
        if self._current_id == tab_id:
            if len(valid_id_values) > 0:
                self._current_id = valid_id_values[0]
                self.setCurrentIndex(self._current_id)
        self.removeTab(tab_id)

    @property
    def model(self):
        tab_id = self.currentIndex()
        try:
            pc = self._context[tab_id]
        except KeyError:
            LOG.error(f"Plotting context is missing for tab {tab_id}")
            LOG.error(self._context)
        else:
            pc._unit_lookup = self._unit_lookup
            return pc

    @property
    def plotter(self):
        tab_id = self.currentIndex()
        try:
            return self._plotter[tab_id]
        except KeyError:
            LOG.error(f"PlotWidget is missing for tab {tab_id}")
            LOG.error(self._plotter)

    @Slot(object)
    def update_plot_details(self, input):
        """This will change the line colour, thickness, etc.
        At the moment it doesn't do anything."""

    @Slot()
    def update_plots(self):
        """This will change the line colour, thickness, etc.
        At the moment it doesn't do anything."""
        for plotter in self._plotter:
            try:
                plotter.plot_data(update_only=True)
            except Exception as e:
                LOG.error(f"Plotting failed: {e}")
